# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Query helper definitions."""
from abc import ABC, abstractmethod
from enum import Enum

from attr import attrs, attrib, Factory

from . utility import export

__all__ = ['KqlQuery']


# Query definition
@attrs
class KqlQuery(object):
    """KqlQuery definition."""

    name = attrib(default=None)
    query = attrib(default=None)
    description = attrib(default=None)
    data_source = attrib(default=None)
    data_families = attrib(default=Factory(list))
    data_environments = attrib(default=Factory(list))
    optional_params = attrib(default=Factory(list))


@export
class DataFamily(Enum):
    """
    Enumeration of data families.

    Used to identify which queries are relevant for which
    data sources.
    """

    WindowsSecurity = 1
    LinuxSecurity = 2
    SecurityAlert = 3

    @classmethod
    def parse(cls, value):
        """
        Conver string or int to enum.

            :param value: value to parse
        """
        if isinstance(value, cls):
            return value

        parsed_enum = None
        if isinstance(value, str):
            try:
                parsed_enum = cls[value]
            except KeyError:
                pass
        if isinstance(value, int):
            parsed_enum = cls(value)
        return parsed_enum


@export
class DataEnvironment(Enum):
    """
    Enumeration of data environments.

    Used to identify which queries are relevant for which
    data sources.
    """

    LogAnalytics = 1
    Kusto = 2

    @classmethod
    def parse(cls, value):
        """
        Conver string or int to enum.

            :param value: value to parse
        """
        if isinstance(value, cls):
            return value

        parsed_enum = None
        if isinstance(value, str):
            try:
                parsed_enum = cls[value]
            except KeyError:
                pass
        if isinstance(value, int):
            parsed_enum = cls(value)
        return parsed_enum


@export
class QueryParamProvider(ABC):
    """
    Abstract type for QueryParamProvider.

    Method query_params must be overridden by derived classes.

    """

    @property
    @abstractmethod
    def query_params(self):
        """
        Return dict of query parameters.

        These parameters are sourced in the object
        implementing this method.

        Returns:
            dict -- dictionary of query parameter values.

        """
        return {}



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
eventcluster module.

Contains:
dbcluster_events: generic clustering method using DBSCAN designed to summarize
process events and other similar data by grouping on common features.

add_process_features: derives numerical features from text features such as
commandline and process path.
"""
from math import log10, floor

import numpy as np
import pandas as pd
from sklearn.cluster import DBSCAN
from sklearn.preprocessing import Normalizer

from .. nbtools.utility import export, pd_version_23
from .. _version import VERSION

__version__ = VERSION
__author__ = 'Ian Hellen'


@export
def dbcluster_events(data, cluster_columns=None, verbose=False, normalize=True,
                     time_column='TimeCreatedUtc',
                     max_cluster_distance=0.01, min_cluster_samples=2,
                     **kwargs):
    """
    Cluster data set according to cluster_columns features.

    Uses sklearn DBSCAN
        :param data: Input data as a pandas DataFrame or numpy array
        :param cluster_columns=None: List of columns to use for features
            for DataFrame this is a list of column names
            for numpy array this is a list of column indexes
        :param verbose=False: Print additional information about clustering
            results
        :param normalize=True: Normalize the input data (should probably always be True)
        :param time_column='TimeCreatedUtc': If there is a time column the output
            data will be ordered by this
        :param max_cluster_distance=0.01: DBSCAN eps (max cluster member distance)
        :param min_cluster_samples=2: DBSCAN min_samples (the minimum cluster size)
        :param **kwargs: Other arguments are passed to DBSCAN constructor
    """
    allowed_types = [np.ndarray, pd.DataFrame]

    x_input = None
    if isinstance(data, pd.DataFrame):
        src_columns = data.columns
        if cluster_columns is None:
            x_input = data.values
        else:
            x_input = data[cluster_columns].values
    elif isinstance(data, np.ndarray):
        if cluster_columns is None:
            x_input = data
        else:
            x_input = data[:, cluster_columns].values

    if x_input is None:
        mssg = 'Input data not in expected format.\n{} is not one of allowed types {}'
        type_list = ', '.join([str(t) for t in allowed_types])
        mssg = mssg.format(str(type(data)), type_list)
        raise ValueError(mssg)

    # Create output frame
    clustered_events = pd.DataFrame(columns=src_columns)

    # Create DBSCAN cluster object
    db_cluster = DBSCAN(eps=max_cluster_distance,
                        min_samples=min_cluster_samples,
                        **kwargs)

    # Normalize the data (most clustering algorithms don't do well with
    # unnormalized data)
    if normalize:
        x_norm = Normalizer().fit_transform(x_input)
    else:
        x_norm = x_input

    # fit the data set
    db_cluster.fit(x_norm)
    labels = db_cluster.labels_
    cluster_set, counts = np.unique(labels, return_counts=True)
    if verbose:
        print('Clustering for set size ', len(x_norm),
              ' - ', len(cluster_set), ' clusters')
        print('Individual cluster sizes: ',
              ', '.join([str(c) for c in counts]))

    # Iterate through clusters, adding exemplar to output frame
    # pylint: disable=C0200
    # we need to know the index of the item within the loop
    for idx in range(len(cluster_set)):
        cluster_id = cluster_set[idx]
        class_members = labels == cluster_id
        if isinstance(data, pd.DataFrame):
            last_event_time = data[class_members][-1:][time_column].iat[0]
        else:
            last_event_time = None

        if cluster_id == -1:
            # 'Noise' events are individual items that could not be assigned
            # to a cluster and so are unique
            if pd_version_23():
                clustered_events = clustered_events.append(
                    data[class_members].assign(Clustered=False,
                                               ClusterId=cluster_id,
                                               ClusterSize=1,
                                               LastEventTime=last_event_time),
                    sort=False)
            else:
                clustered_events = clustered_events.append(
                    data[class_members].assign(Clustered=False,
                                               ClusterId=cluster_id,
                                               ClusterSize=1,
                                               LastEventTime=last_event_time))
        else:
            # Otherwise, just choose the first example of the cluster set
            if pd_version_23():
                clustered_events = clustered_events.append(
                    data[class_members].assign(Clustered=True,
                                               ClusterId=cluster_id,
                                               ClusterSize=counts[idx],
                                               LastEventTime=last_event_time)[0:1],
                    sort=False)
            else:
                clustered_events = clustered_events.append(
                    data[class_members].assign(Clustered=True,
                                               ClusterId=cluster_id,
                                               ClusterSize=counts[idx],
                                               LastEventTime=last_event_time)[0:1])

    # pylint: enable=C0200
    if verbose:
        print('Cluster output rows: ', len(clustered_events))

    return clustered_events, db_cluster, x_norm


@export
def add_process_features(input_frame, path_separator=None, force=False):
    r"""
    Add numerical features based on patterns of command line and process name.

        :param input_frame: The input dataframe
        :param path_separator=None: Path separator - if not supplied, try to determine
            from 'NewProcessName' column of first 10 rows
        :param force=False: Forces re-calculation of feature columns even if they
            already exist

    Features:
        processNameLen: length of process file name (inc path)
        processNameTokens: the number of elements in the path
        processName: the process file name (minus path)
        commandlineTokens: number of space-separated tokens in the command line
        commandlineLen: length of the command line
        commandlineLogLen: log10 length of commandline
        isSystemSession: 1 if session Id is 0x3e7 for Windows or -1 for Linux
        commandlineTokensFull: counts number of token separators in commandline
            [\s\-\\/\.,"\'|&:;%$()]
        pathScore: sum of ord() value of characters in path
        pathLogScore: log10 of pathScore
        commandlineScore: sum of ord() value of characters in commandline
        commandlineLogScore: log10 of commandlineScore
    """
    output_df = input_frame.copy()

    # Set any NaN values to empty string
    if 'NewProcessName' in output_df and 'CommandLine' in output_df:
        output_df[['NewProcessName', 'CommandLine']] = output_df[['NewProcessName',
                                                                  'CommandLine']].fillna(value='')

    # try to determine the path separator
    if path_separator is None:
        sample_df = output_df.head(10)
        lx_path = len(sample_df[sample_df['NewProcessName'].str.contains('/')])
        if lx_path:
            path_separator = '/'
        else:
            path_separator = '\\'

    # Create features from process name and command line
    if 'NewProcessName' in output_df:
        if 'processNameLen' not in output_df or force:
            output_df['processNameLen'] = output_df.apply(lambda x:
                                                          len(x.NewProcessName),
                                                          axis=1)
        if 'processNameTokens' not in output_df or force:
            output_df['processNameTokens'] = output_df.apply(lambda x:
                                                             len(x.NewProcessName.split(
                                                                 path_separator)),
                                                             axis=1)
        if 'processName' not in output_df or force:
            output_df['processName'] = output_df.apply(lambda x:
                                                       x.NewProcessName.split(
                                                           path_separator)[-1],
                                                       axis=1)
        if 'pathScore' not in output_df or force:
            output_df['pathScore'] = output_df.apply(lambda x:
                                                     _string_score(
                                                         x.NewProcessName),
                                                     axis=1)
        if 'pathLogScore' not in output_df or force:
            output_df['pathLogScore'] = output_df.apply(lambda x:
                                                        log10(x.pathScore)
                                                        if x.pathScore else 0,
                                                        axis=1)

    if 'CommandLine' in output_df:
        if 'commandlineTokens' not in output_df or force:
            output_df['commandlineTokens'] = output_df.apply(lambda x:
                                                             len(x.CommandLine.split(
                                                                 path_separator)),
                                                             axis=1)
        if 'commandlineLen' not in output_df or force:
            output_df['commandlineLen'] = output_df.apply(lambda x:
                                                          len(x.CommandLine),
                                                          axis=1)
        if 'commandlineLogLen' not in output_df or force:
            output_df['commandlineLogLen'] = output_df.apply(lambda x:
                                                             log10(
                                                                 x.commandlineLen)
                                                             if x.commandlineLen else 0, axis=1)
        if 'commandlineTokensFull' not in output_df or force:
            delim_rgx = r'[\s\-\\/\.,"\'|&:;%$()]'
            output_df['commandlineTokensFull'] = (output_df[['CommandLine']]
                                                  .apply(lambda x: x.str.count(delim_rgx),
                                                         axis=1))

        if 'commandlineScore' not in output_df or force:
            output_df['commandlineScore'] = output_df.apply(lambda x:
                                                            _string_score(
                                                                x.CommandLine),
                                                            axis=1)
        if 'commandlineLogScore' not in output_df or force:
            output_df['commandlineLogScore'] = output_df.apply(lambda x:
                                                               log10(
                                                                   x.commandlineScore)
                                                               if x.commandlineScore else 0,
                                                               axis=1)

    if 'SubjectLogonId' in output_df:
        if (('isSystemSession' not in output_df or force) and 'SubjectLogonId' in output_df):
            output_df['isSystemSession'] = output_df.apply(lambda x:
                                                           True if x.SubjectLogonId == '0x3e7' or
                                                           x.SubjectLogonId == '-1' else False,
                                                           axis=1)

    return output_df


@export
def delim_count(input_row: pd.Series, column: str,
                delim_list: str = r'[\s\-\\/\.,"\'|&:;%$()]') -> int:
    r"""
    Count the delimiters in input column.

        :param input_row:pd.Series: The series to process
        :param column:str: Column name
        :param delim_list:str=r'[\s\-\\/\."\'|&:;%$()]: delimiters to use.
    """
    return input_row[column].str.count(delim_list)


@export
def char_ord_score(input_row: pd.Series, column: str, scale: int = 1) -> int:
    """
    Return sum of ord values of characters in string.

    This function sums the ordinal value of each character in the
    input string. Two strings with minor differences will result in
    a similar score. However, for strings with highly variable content
    (e.g. command lines or http requests containing GUIDs) this may result
    in too much variance to be useful when you are trying to detect
    similar patterns. You can use the scale parameter to reduce the
    influence of features using this function on clustering and anomaly
    algorithms.

    Arguments:
        input_row {pd.Series} -- The series to process
        column {str} -- Column name
        scale {int} -- reduce the scale of the feature (reducing the
            influence of variations this feature on the )

    Keyword Arguments:
        delimiter {str} -- Delimiter used to split the column string (default: {' '})

    Returns:
        {int} -- count of tokens

    """
    return floor(sum([ord(x) for x in input_row[column]]) / scale)


@export
def token_count(input_row: pd.Series, column: str, delimiter: str = ' ') -> int:
    """
    Return delimiter-separated tokens pd.Series column.

    Arguments:
        input_row {pd.Series} -- The series to process
        column {str} -- Column name

    Keyword Arguments:
        delimiter {str} -- Delimiter used to split the column string (default: {' '})

    Returns:
        {int} -- count of tokens

    """
    return len(input_row[column].split(delimiter))


def _string_score(input_str):
    """Sum the ord(c) for characters in a string."""
    return sum([ord(x) for x in input_str])



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""query_schema test class."""
import unittest
import json

# from .. nbtools import query_builtin_queries as queries
# from .. nbtools.query_mgr import replace_query_params, add_query
# from .. nbtools.query_defns import DataFamily, DataEnvironment, KqlQuery
from .. msticpy.nbtools.entityschema import Entity, Account, Host, Process, File, IpAddress


class Testnbtools(unittest.TestCase):
    """Unit test class."""

    def entity_creation(self):
        try:
            file = './msticpy/tests/testdata/entities.json'
            with open(file, 'r') as file_handle:
                txt = file_handle.read()
                entity_dict = json.loads(txt)

            parsed_entities = []
            for _, entity in entity_dict.items():
                e = Entity.instantiate_entity(entity)
                self.assertIsInstance(e, Entity)

                if e['Type'] == 'account':
                    self.assertIsInstance(e, Account)
                    self.assertTrue('Name' in e)
                    self.assertGreater(len(e.Name), 0)
                elif e['Type'] == 'host':
                    self.assertIsInstance(e, Host)
                    self.assertTrue('HostName' in e)
                    self.assertGreater(len(e.HostName), 0)
                elif e['Type'] == 'process':
                    self.assertIsInstance(e, Process)
                    self.assertTrue('ProcessId' in e)
                    self.assertGreater(len(e.ProcessId), 0)
                elif e['Type'] == 'file':
                    self.assertIsInstance(e, File)
                    self.assertTrue('Name' in e)
                    self.assertGreater(len(e.Name), 0)
                elif e['Type'] == 'ipaddress':
                    self.assertIsInstance(e, IpAddress)
                    self.assertTrue('Address' in e)
                    self.assertGreater(len(e.Address), 0)

                parsed_entities.append(e)

            self.assertGreaterEqual(len(parsed_entities), 7)

        except Exception as ex:
            self.fail(msg='Exception {}'.format(str(ex)))


if __name__ == '__main__':
    unittest.main()
    print('bye')



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for IoCExtract class."""

import re
from collections import namedtuple, defaultdict
from urllib.parse import unquote

import pandas as pd
from .. nbtools.utility import export
from .. _version import VERSION

__version__ = VERSION
__author__ = 'Ian Hellen'


def _compile_regex(regex):

    return re.compile(regex, re.I | re.X | re.M)

IoCPattern = namedtuple('IoCPattern', ['ioc_type', 'comp_regex', 'priority'])


@export
class IoCExtract(object):
    """
    IoC Extractor - looks for common IoC patterns in input strings.

    The extract() method takes either a string or a pandas DataFrame
    as input. When using the string option as an input extract will
    return a dictionary of results. When using a DataFrame the results
    will be returned as a new DataFrame with the following columns:
    IoCType: the mnemonic used to distinguish different IoC Types
    Observable: the actual value of the observable
    SourceIndex: the index of the row in the input DataFrame from
    which the source for the IoC observable was extracted.

    The class has a number of built-in IoC regex definitions.
    These can be retrieved using the ioc_types attribute.

    Addition IoC definitions can be added using the add_ioc_type
    method.

    Note: due to some ambiguity in the regular expression patterns
    for different types and observable may be returned assigned to
    multiple observable types. E.g. 192.168.0.1 is a also a legal file
    name in both Linux and Windows. Linux file names have a particularly
    large scope in terms of legal characters so it will be quite common
    to see other IoC observables (or parts of them) returned as a
    possible linux path.
    """

    IPV4_REGEX = r'(?P<ipaddress>(?:[0-9]{1,3}\.){3}[0-9]{1,3})'
    IPV6_REGEX = r'(?<![:.\w])(?:[A-F0-9]{1,4}:){7}[A-F0-9]{1,4}(?![:.\w])'
    DNS_REGEX = r'((?=[a-z0-9-]{1,63}\.)[a-z0-9]+(-[a-z0-9]+)*\.){2,}[a-z]{2,63}'
    # dns_regex = '\\b((?=[a-z0-9-]{1,63}\\.)[a-z0-9]+(-[a-z0-9]+)*\\.){2,}[a-z]{2,63}\\b'

    URL_REGEX = r'''
            (?P<protocol>(https?|ftp|telnet|ldap|file)://)
            (?P<userinfo>([a-z0-9-._~!$&\'()*+,;=:]|%[0-9A-F]{2})*@)?
            (?P<host>([a-z0-9-._~!$&\'()*+,;=]|%[0-9A-F]{2})*)
            (:(?P<port>\d*))?
            (/(?P<path>([^?\# ]|%[0-9A-F]{2})*/?))?
            (\?(?P<query>([a-z0-9-._~!$&'()*+,;=:/?@]|%[0-9A-F]{2})*))?
            (\#(?P<fragment>([a-z0-9-._~!$&'()*+,;=:/?@]|%[0-9A-F]{2})*))?'''

    WINPATH_REGEX = r'''
            (?P<root>[a-z]:|\\\\[a-z0-9_.$-]+||[.]+)
            (?P<folder>\\(?:[^\/:*?"\'<>|\r\n]+\\)*)
            (?P<file>[^\\/*?""<>|\r\n ]+)'''
    # Linux simplified - this ignores some legal linux paths avoid matching too much
    # TODO - also matches URLs!
    LXPATH_REGEX = r'''(?P<root>/+||[.]+)
            (?P<folder>/(?:[^\\/:*?<>|\r\n]+/)*)
            (?P<file>[^/\0<>|\r\n ]+)'''

    MD5_REGEX = r'(?:^|[^A-Fa-f0-9])(?P<hash>[A-Fa-f0-9]{32})(?:$|[^A-Fa-f0-9])'
    SHA1_REGEX = r'(?:^|[^A-Fa-f0-9])(?P<hash>[A-Fa-f0-9]{40})(?:$|[^A-Fa-f0-9])'
    SHA256_REGEX = r'(?:^|[^A-Fa-f0-9])(?P<hash>[A-Fa-f0-9]{64})(?:$|[^A-Fa-f0-9])'

    _content_regex = {}

    def __init__(self):
        """Intialize new instance of IoCExtract."""
        # IP Addresses
        self.add_ioc_type('ipv4', self.IPV4_REGEX, 0)
        self.add_ioc_type('ipv6', self.IPV6_REGEX, 0)

        # Dns Domains
        # TODO - This also matches IP addresses
        self.add_ioc_type('dns', self.DNS_REGEX, 1)

        # Http requests
        self.add_ioc_type('url', self.URL_REGEX, 0)

        # File paths
        # Windows
        self.add_ioc_type('windows_path', self.WINPATH_REGEX, 2)

        self.add_ioc_type('linux_path', self.LXPATH_REGEX, 2)

        # MD5, SHA1, SHA256 hashes
        self.add_ioc_type('md5_hash', self.MD5_REGEX, 1)
        self.add_ioc_type('sha1_hash', self.SHA1_REGEX, 1)
        self.add_ioc_type('sha256_hash', self.SHA256_REGEX, 1)

    # Public members

    def add_ioc_type(self, ioc_type: str, ioc_regex: str, priority: int = 0):
        """
        Add an IoC type and regular expression to use to the built-in set.

        Note: adding an ioc_type that exists in the internal set will overwrite that item
        Regular expressions are compiled with re.I | re.X | re.M (Ignore case, Verbose
        and MultiLine)
            :param: ioc_type - a unique name for the IoC type
            :param: ioc_regex - a regular expression used to search for the type
            :type ioc_type: str
            :type ioc_regex: str
        """
        if ioc_type is None or ioc_type.strip() is None:
            raise Exception('No value supplied for ioc_type parameter')
        if ioc_regex is None or ioc_regex.strip() is None:
            raise Exception('No value supplied for ioc_regex parameter')

        self._content_regex[ioc_type] = IoCPattern(ioc_type=ioc_type,
                                                   comp_regex=_compile_regex(regex=ioc_regex),
                                                   priority=priority)

    @property
    def ioc_types(self) -> dict:
        """
        Return the current set of IoC types and regular expressions.

            :rtype: dict of IoC Type names and regular expressions
        """
        return self._content_regex

    def extract(self, src: str = None, data: pd.DataFrame = None,
                columns: list = None, os_family='Windows',
                ioc_types: list = None):
        """
        Extract IoCs from either a string or pandas DataFrame.

        Keyword Arguments:
            src {str} --  source string in which to look for IoC patterns
                (default: {None})
            data {pd.DataFrame} -- input DataFrame from which to read source strings
                 (default: {None})
            columns {list} -- The list of columns to use as source strings,
                if the data parameter is used.  (default: {None})
            os_family {str} -- 'Linux' or 'Windows' (default: {'Windows'})
            ioc_types {list({str})} -- Restrict matching to just specified
                types (default: {None})


        Returns:
            dict of found observables (if input is a string) or
            DataFrame of observables

        Extract takes either a string or a pandas DataFrame as input.
        When using the string option as an input extract will
        return a dictionary of results.
        When using a DataFrame the results will be returned as a new
        DataFrame with the following columns:
        - IoCType: the mnemonic used to distinguish different IoC Types
        - Observable: the actual value of the observable
        - SourceIndex: the index of the row in the input DataFrame from
        which the source for the IoC observable was extracted.

        """
        if src and src.strip():
            return self._scan_for_iocs(src, os_family)

        if data is None:
            raise Exception('No source data was supplied to extract')

        # Handle DataFrame option
        assert isinstance(data, pd.DataFrame)

        if columns is None:
            raise Exception(
                'No values where supplied for the columns parameter')

        col_set = set(columns)
        if not col_set <= set(data.columns):
            missing_cols = [elem for elem in col_set if elem not in data.colums]
            raise Exception('Source column(s) {} not found in supplied DataFrame'
                            .format(', '.join(missing_cols)))

        result_columns = ['IoCType', 'Observable', 'SourceIndex']
        result_frame = pd.DataFrame(columns=result_columns)
        for idx, datarow in data.iterrows():
            for col in columns:
                ioc_results = self._scan_for_iocs(datarow[col], os_family, ioc_types)
                for result_type, result_set in ioc_results.items():
                    if result_set:
                        for observable in result_set:
                            result_row = pd.Series(
                                data=[result_type, observable, idx], index=result_columns)
                            result_frame = result_frame.append(
                                result_row, ignore_index=True)

        return result_frame

    def validate(self, input_str: str, ioc_type: str) -> bool:
        """
        Return true if the input_str matches the corresponding regex.

        Arguments:
            :input_str str: the string to test
            :ioc_type str: the regex pattern to use

        Returns:
            bool - True if match.

        """
        if ioc_type not in self._content_regex:
            raise KeyError('Unknown type {}. Valid types are: {}'
                           .format(ioc_type, list(self._content_regex.keys())))
        rgx = self._content_regex[ioc_type]
        return rgx.comp_regex.fullmatch(input_str) is not None

    # Private methods
    def _scan_for_iocs(self, src: str, os_family: str, ioc_types: list = None) -> dict:
        """Return IoCs found in the string."""
        ioc_results = defaultdict(set)
        iocs_found = {}

        for (ioc_type, rgx_def) in self._content_regex.items():
            if ioc_types and ioc_type not in ioc_types:
                continue

            if os_family == 'Linux' and rgx_def.ioc_type == 'windows_path':
                continue
            elif os_family == 'Windows' and rgx_def.ioc_type == 'linux_path':
                continue

            match_pos = 0
            for rgx_match in rgx_def.comp_regex.finditer(src, match_pos):
                if rgx_match is not None:
                    self._add_highest_pri_match(iocs_found,
                                                rgx_match.group(),
                                                rgx_def)
                    if ioc_type == 'url':
                        decoded_url = unquote(rgx_match.group())
                        for url_match in rgx_def.comp_regex.finditer(decoded_url, match_pos):
                            if url_match is not None:
                                self._add_highest_pri_match(iocs_found,
                                                            url_match.group(),
                                                            rgx_def)
                                self._add_highest_pri_match(iocs_found,
                                                            url_match.groupdict()['host'],
                                                            self._content_regex['dns'])
                    match_pos = rgx_match.end()
                else:
                    break
        for ioc, ioc_result in iocs_found.items():
            ioc_results[ioc_result[0]].add(ioc)

        return ioc_results

    def _add_highest_pri_match(self, iocs_found: dict, current_match: str, current_def: IoCPattern):
        # if we already found a match for this item and the previous
        # ioc type is more specific then don't add this to the results
        if current_match in iocs_found and current_def.priority > iocs_found[current_match][1]:
            return
        else:
            iocs_found[current_match] = (current_def.ioc_type, current_def.priority)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
security_alert_graph.

Creates an entity graph for the alert.
"""
import networkx as nx
import pandas as pd

from . security_alert import SecurityAlert
from . entityschema import Entity
from . utility import export, is_not_empty
from .. _version import VERSION

__version__ = VERSION
__author__ = 'Ian Hellen'


@export
def create_alert_graph(alert: SecurityAlert):
    """Create a networkx graph from the alert and contained entities."""
    alertentity_graph = nx.Graph(id='AlertGraph')

    alertentity_graph.add_node(alert['AlertType'],
                               name=alert['AlertType'],
                               time=str(alert['StartTimeUtc']),
                               description='Alert: ' +
                               alert['AlertDisplayName'],
                               color='red',
                               node_type='alert')

    os_family = alert.os_family

    # Cycle through entities
    for entity in alert.entities:
        (e_name, e_desc) = _get_name_and_description(entity, os_family)

        alertentity_graph.add_node(e_name, entitytype=entity['Type'], name=e_name,
                                   description=e_desc, color='green',
                                   node_type='entity', source=str(entity))

        # add an edge by default to the alert
        alertentity_graph.add_edge(alert['AlertType'], e_name)

        # Rather than just add edges to the alert, we want to follow the 'natural'
        # relationships between entities and child entities
        # So if this entity has a property that is an entity, we add an edge to it
        # and prune any edge that it might have to the alert
        if isinstance(entity, Entity):
            ent_props = entity.properties
        elif isinstance(entity, dict):
            ent_props = entity
        else:
            continue
        for prop, rel_entity in [(p, v) for (p, v) in ent_props.items()
                                 if isinstance(v, Entity)]:
            if rel_entity['Type'] == 'host':
                # don't add a new edge to the host
                continue

            # get the node id of the related entity and add an edge if it
            # doesn't already exist
            (related_entity, _) = _get_name_and_description(rel_entity)
            if not alertentity_graph.has_edge(related_entity, e_name):
                alertentity_graph.add_edge(e_name, related_entity, description=prop,
                                           color='green', weight=1,
                                           line_type='SHORT_DASH')

            # if we have a previously created an edge to the alert, remove it
            if alertentity_graph.has_edge(alert['AlertType'], related_entity):
                alertentity_graph.remove_edge(
                    alert['AlertType'], related_entity)

        # if we haven't added an edge to this entity from anything else,
        # add one to the alert
        if not alertentity_graph.neighbors(e_name):
            alertentity_graph.add_edge(alert['AlertType'], e_name)

    return alertentity_graph


@export
def add_related_alerts(related_alerts: pd.DataFrame, alertgraph: nx.Graph) ->nx.Graph:
    """
    Add related alerts to the graph.

    Link to the entity that is common to both alerts.
    """
    related_alerts_graph = alertgraph.copy()

    alert_host_node = _find_graph_node(related_alerts_graph, 'host', '')

    related_alerts.apply(lambda x: _add_alert_node(
        related_alerts_graph, x), axis=1)
    if alert_host_node:
        related_alerts.apply(lambda x:
                             _add_related_alert_edges(related_alerts_graph,
                                                      x,
                                                      alert_host_node), axis=1)
    return related_alerts_graph


def _add_related_alert_edges(related_alerts_graph, alert_row, default_node):
    related_alert = SecurityAlert(alert_row)
    if related_alert.primary_account is not None:
        acct_node = _find_graph_node(related_alerts_graph, 'account',
                                     related_alert.primary_account.qualified_name)
        if acct_node is not None:
            _add_related_alert_edge(
                related_alerts_graph, acct_node, related_alert)

    if related_alert.primary_process is not None:
        proc_node = _find_graph_node(related_alerts_graph,
                                     'process',
                                     related_alert.primary_process.ProcessFilePath)
        if proc_node is not None:
            _add_related_alert_edge(
                related_alerts_graph, proc_node, related_alert)

    if related_alert.primary_host is not None:
        host_node = _find_graph_node(related_alerts_graph,
                                     'host', related_alert.primary_host['HostName'])
        if host_node is not None:
            _add_related_alert_edge(
                related_alerts_graph, host_node, related_alert)

    # if we haven't added an edge to this entity from anything else,
    # add one to the alert
    if not related_alerts_graph[related_alert['AlertType'] + '(R)']:
        _add_related_alert_edge(related_alerts_graph,
                                default_node, related_alert)


def _add_alert_node(nx_graph, alert):
    """Add alert node to the graph."""
    nx_graph.add_node(alert['AlertType'] + '(R)',
                      name=alert['AlertType'],
                      time=str(alert['StartTimeUtc']),
                      displayname=alert['AlertDisplayName'],
                      color='red',
                      count=0,
                      node_type='alert')


def _find_graph_node(nx_graph, node_type, target_name):
    """Find a node with a given name and type."""
    node_prefix = '{}: {}'.format(node_type, target_name)
    nodes = [n for (n, n_type) in
             nx.get_node_attributes(nx_graph, 'entitytype').items()
             if n_type == node_type and n.startswith(node_prefix)]
    if nodes:
        return nodes[0]


def _add_related_alert_edge(nx_graph, source, target):
    """Add related alert to an existing graph."""
    count_attrs = nx.get_node_attributes(nx_graph, 'count')
    target_node = target['AlertType'] + '(R)'
    if target_node in count_attrs:
        current_count = count_attrs[target_node]
    else:
        current_count = 0
    current_count += 1

    description = 'Related alert: {}  Count:{}'.format(target['AlertType'],
                                                       current_count)
    node_attrs = {target_node: {
        'count': current_count, 'description': description}}
    nx.set_node_attributes(nx_graph, node_attrs)
    nx_graph.add_edge(source, target_node, weight=0.7,
                      description='Related Alert')


def _get_account_qualified_name(account):
    if 'Name' in account:
        name = account['Name']
    if 'NTDomain' in account:
        return '{}\\{}'.format(account['NTDomain'], name)
    else:
        return name


def _get_name_and_description(entity, os_family='Windows'):
    """Get name and description for entity."""
    e_name = None
    e_description = None

    if entity['Type'] == 'host':
        e_name, e_description = _get_host_name_desc(entity, os_family)
    elif entity['Type'] == 'account':
        e_name, e_description = _get_account_name_desc(entity)
    elif entity['Type'] == 'host-logon-session':
        e_name = 'host-logon-session'
        e_description = 'Logon session {}\n(Start time: {})'.format(entity['SessionId'],
                                                                    entity['StartTimeUtc'])
    elif entity['Type'] == 'process':
        e_name, e_description = _get_process_name_desc(entity)
    elif entity['Type'] == 'file':
        e_name, e_description = _get_file_name_desc(entity)
    elif entity['Type'] == 'ip' or entity['Type'] == 'ipaddress':
        e_name, e_description = _get_ip_name_desc(entity)
    elif entity['Type'] == 'dns':
        e_name = '{}: {}'.format(entity['Type'], entity['DomainName'])
        e_description = e_name
    else:
        # Any other type of entity
        e_name, e_description = _get_other_name_desc(entity)

    return e_name, e_description


# Methods to construct name and description
def _get_other_name_desc(entity):
    if 'Name' in entity:
        e_name = entity['Name']
        e_name = '{}: {}'.format(entity['Type'], e_name)
    else:
        e_name = entity['Type']

    if isinstance(entity, Entity):
        ent_props = entity.properties
    elif isinstance(entity, dict):
        ent_props = entity
    else:
        ent_props = {'unknown': None}

    # Nasty dict comprehension to join all other items in the dictionary into a string
    e_properties = '\n'.join({'{}:{}'.format(k, v) for (k, v)
                              in ent_props.items() if (k not in ('Type', 'Name') and
                                                       isinstance(v, str))})
    e_description = '{}\n{})'.format(e_name, e_properties)
    return e_name, e_description


def _get_ip_name_desc(entity):
    e_name = entity['Address']
    e_name = '{}: {}'.format(entity['Type'], e_name)
    if 'Location' in entity and entity['Location']:
        e_description = '{}\nc={}, st={}, city={}'.format(
            e_name,
            entity['Location'].CountryCode,
            entity['Location'].State,
            entity['Location'].City)
    else:
        e_description = e_name
    return e_name, e_description


def _get_file_name_desc(entity):
    e_name = entity['FullPath']
    e_name = '{}: {}'.format(entity['Type'], e_name)
    e_description = e_name
    return e_name, e_description


def _get_process_name_desc(entity):
    if 'ProcessFilePath' in entity:
        path = entity.ProcessFilePath
    elif 'ImageFile' in entity and entity['ImageFile']:
        path = entity['ImageFile']['FullPath']
    else:
        path = 'unknown'
    pid = entity.ProcessId if entity.ProcessId else 'PID unknown'
    e_name = path + ' [' + pid + ']'
    e_name = '{}: {}'.format(entity['Type'], e_name)
    e_description = '{}\n(cmdline: \'{}\')'.format(e_name, entity.CommandLine)
    return e_name, e_description


def _get_account_name_desc(entity):
    e_name = (entity['NTDomain'] +
              '\\' if 'NTDomain' in entity else '') + entity['Name']
    e_name = '{}: {}'.format(entity['Type'], e_name)
    if 'IsDomainJoined' in entity:
        domain_joined = entity['IsDomainJoined']
    else:
        domain_joined = 'false'
    if 'LogonId' in entity:
        e_description = '{}\n(LogonId: {}, Domain-joined: {})'.format(e_name,
                                                                      entity.LogonId,
                                                                      domain_joined)
    else:
        e_description = '{}\n(Domain-joined: {})'.format(e_name, domain_joined)
    return e_name, e_description


def _get_host_name_desc(entity, os_family):
    if 'DnsDomain' in entity and is_not_empty(entity['DnsDomain']):
        e_name = '{}.{}'.format(entity['HostName'], entity['DnsDomain'])
    elif 'NTDomain' in entity and is_not_empty(entity['NTDomain']):
        e_name = '{}/{}'.format(entity['NTDomain'], entity['HostName'])
    else:
        e_name = entity['HostName']
    e_name = '{}: {}'.format(entity['Type'], e_name)

    if 'IsDomainJoined' in entity:
        domain_joined = entity['IsDomainJoined']
    else:
        domain_joined = 'false'
    if 'OSFamily' in entity:
        os_family = entity['OSFamily']
    e_description = '{}\n({}, Domain-joined: {})'.format(e_name,
                                                         os_family, domain_joined)

    return e_name, e_description



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for SecurityAlert class."""
import json
from json import JSONDecodeError

import pandas as pd

from .._version import VERSION
from .entityschema import Entity, UnknownEntity
from .security_base import SecurityBase
from .utility import export

__version__ = VERSION
__author__ = 'Ian Hellen'


@export
class SecurityAlert(SecurityBase):
    """
    Security Alert Class.

    Instantiates a security alert from a pandas Series and provides
    convenience access methods to retrieve properties.
    """

    def __init__(self, src_row: pd.Series = None):
        """Instantiate a security alert from a pandas Series."""
        super().__init__(src_row=src_row)

        # add entities to dictionary to remove dups
        self._src_entities = dict()
        if 'Entities' in src_row:
            self._extract_entities(src_row)

        if 'ExtendedProperties' in src_row:
            if isinstance(src_row.ExtendedProperties, dict):
                self.extended_properties = src_row.ExtendedProperties
            elif isinstance(src_row.ExtendedProperties, str):
                try:
                    self.extended_properties = json.loads(src_row.ExtendedProperties)
                except JSONDecodeError:
                    pass
        else:
            self.extended_properties = []
        self._find_os_family()

    @property
    def entities(self) -> list:
        """Return a list of the Security Alert entities."""
        return list(self._src_entities.values())

    @property
    def query_params(self) -> dict:
        """
        Query parameters derived from alert.

        Returns:
            dict(str, str) -- Dictionary of parameter names

        """
        params_dict = super().query_params
        if ('system_alert_id' not in params_dict or
                params_dict['system_alert_id'] is None):
            params_dict['system_alert_id'] = self._ids['SystemAlertId']
        return params_dict

    def to_html(self, show_entities=False):
        """Return the item as HTML string."""
        title = '''
        <h3>Alert: '{name}'</h3><br>time=<b>{start}</b>, entity=<b>{entity}</b>, id=<b>{id}</b>
        <br/>
        '''.format(start=self._source_data['StartTimeUtc'],
                   name=self._source_data['AlertDisplayName'],
                   entity=self._source_data.get(
                       'CompromisedEntity', 'unknown'),
                   id=self._source_data['SystemAlertId'])
        return title + super().to_html(show_entities)

    # Public methods
    def __str__(self):
        """Print the alert contents to stdout."""
        alert_props = str(super().__str__())

        if self.extended_properties:
            str_rep = [f'ExtProp: {prop}: {val}' for prop, val in
                       self.extended_properties.items()]
            alert_props = alert_props + '\n' + '\n'.join(str_rep)

        return alert_props

    # Private methods
    def _resolve_entity_refs(self):
        """
        Resolve and replace entity properties that are '$ref' type.

        When serialized the nested entities can be references to other referenced
        objects in the collection. This iterates through the raw entities and
        replaces referenced items with the actual object reference.
        If the Id referenced by this property exists in the entities dictionary
        we replace the original property with a reference to the entity in the dictionary
        """
        for _, entity in self._src_entities.items():
            if not isinstance(entity, Entity):
                continue
            for prop_name, prop_val in entity.properties.items():
                if isinstance(prop_val, dict) and '$ref' in prop_val:
                    entity_id = prop_val['$ref']
                    if entity_id in self._src_entities:
                        entity[prop_name] = self._src_entities[entity_id]

    def _find_os_family(self):
        """Work out which OSFamily and path separator to use from entities or file paths."""
        self.path_separator = '\\'
        self.os_family = 'Windows'

        # Use OSFamily if any entities have this property set
        os_family_entities = [e for e in self.entities if 'OSFamily' in e]
        if os_family_entities:
            for os_entity in os_family_entities:
                if os_entity['OSFamily'] == 'Linux':
                    self.os_family = 'Linux'
                    self.path_separator = '/'
                    break
        else:
            # Otherwise try to infer from the file paths
            files = [e for e in self.entities if e['Type'] == 'file']
            if files:
                for file in files:
                    if 'Directory' in file and '/' in file['Directory']:
                        self.os_family = 'Linux'
                        self.path_separator = '/'
                        break
            else:
                for proc in [e for e in self.entities
                             if e['Type'] == 'process' and 'ImageFile' in e]:
                    file = proc['ImageFile']
                    if 'Directory' in file and '/' in file['Directory']:
                        self.os_family = 'Linux'
                        self.path_separator = '/'
                        break

    def _extract_entities(self, src_row):
        input_entities = []
        if isinstance(src_row.Entities, str):
            try:
                input_entities = json.loads(src_row['Entities'])
            except json.JSONDecodeError:
                pass
        elif isinstance(src_row.Entities, list):
            input_entities = src_row.Entities

        for ent in input_entities:
            try:
                entity = Entity.instantiate_entity(ent)
            except TypeError:
                # if we didn't instantiate a known entity
                # just add it as it is
                entity = UnknownEntity(**ent)
            if '$id' in ent:
                self._src_entities[ent['$id']] = entity

        self._resolve_entity_refs()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
eventschema.

Module for DataSchema class
"""
from . query_defns import DataFamily, DataEnvironment
from . utility import export
from .. _version import VERSION

__version__ = VERSION
__author__ = 'Ian Hellen'


# TODO Refactor to simpler data structure
@export
class DataSchema:
    """DataSchema class for Log Analytics Queries."""

    DATA_MAPPINGS = {DataEnvironment.LogAnalytics: {},
                     DataEnvironment.Kusto: {}}
    DATA_MAPPINGS[DataEnvironment.LogAnalytics] = {DataFamily.WindowsSecurity: {},
                                                   DataFamily.LinuxSecurity: {},
                                                   DataFamily.SecurityAlert: {}}

    _SECURITY_ALERT = {'table': 'SecurityAlert',
                       'query_project': '''| project
                                            TenantId,
                                            StartTimeUtc = StartTime,
                                            EndTimeUtc = EndTime,
                                            ProviderAlertId = VendorOriginalId,
                                            SystemAlertId,
                                            ProviderName,
                                            VendorName,
                                            AlertType = AlertName,
                                            AlertName,
                                            AlertDisplayName = DisplayName,
                                            Description,
                                            Severity = AlertSeverity,
                                            IsIncident,
                                            ExtendedProperties,
                                            Entities,
                                            ConfidenceLevel,
                                            ConfidenceScore,
                                            ExtendedLinks,
                                            WorkspaceSubscriptionId,
                                            WorkspaceResourceGroup,
                                            TimeGenerated,
                                            ResourceId,
                                            SourceComputerId '''}

    _PROC_CREATE_WIN = {'table': 'SecurityEvent | where EventID == 4688',
                        'query_project': '''| project
                                            TenantId,
                                            Account,
                                            EventID,
                                            TimeGenerated,
                                            Computer,
                                            SubjectUserSid,
                                            SubjectUserName,
                                            SubjectDomainName,
                                            SubjectLogonId,
                                            NewProcessId,
                                            NewProcessName,
                                            TokenElevationType,
                                            ProcessId,
                                            CommandLine,
                                            ParentProcessName,
                                            TargetLogonId,
                                            SourceComputerId
                                            | extend TimeCreatedUtc=TimeGenerated '''}

    _ACCOUNT_LOGON_WIN = {'table': 'SecurityEvent | where EventID == 4624',
                          'query_project': '''| project
                                            TenantId,
                                            Account,
                                            EventID,
                                            TimeGenerated,
                                            SourceComputerId,
                                            Computer,
                                            SubjectUserName,
                                            SubjectDomainName,
                                            SubjectUserSid,
                                            TargetUserName,
                                            TargetDomainName,
                                            TargetUserSid,
                                            TargetLogonId,
                                            LogonProcessName,
                                            LogonType,
                                            AuthenticationPackageName,
                                            Status,
                                            IpAddress,
                                            WorkstationName'''}

    _ACCOUNT_LOGON_FAIL_WIN = {'table': 'SecurityEvent | where EventID == 4625',
                               'query_project': '''| project
                                            TenantId,
                                            Account,
                                            EventID,
                                            TimeGenerated,
                                            SourceComputerId,
                                            Computer,
                                            SubjectUserName,
                                            SubjectDomainName,
                                            SubjectUserSid,
                                            TargetUserName,
                                            TargetDomainName,
                                            TargetUserSid,
                                            TargetLogonId,
                                            LogonProcessName,
                                            LogonType,
                                            AuthenticationPackageName,
                                            Status,
                                            IpAddress,
                                            WorkstationName'''}

    _PROC_CREATE_LX = {'table': 'LinuxAuditD | where EventID == 14688',
                       'query_project': '''| project
                                            TenantId,
                                            Account,
                                            EventID,
                                            TimeGenerated,
                                            SourceComputerId
                                            Computer=node,
                                            SubjectUserSid=uid,
                                            SubjectUserName=user,
                                            SubjectDomainName,
                                            SubjectLogonId=ses,
                                            NewProcessId=pid,
                                            NewProcessName=exe,
                                            TokenElevationType,
                                            ProcessId=ppid,
                                            CommandLine=cmdline,
                                            ParentProcessName,
                                            TargetLogonId,
                                            success,
                                            audit_user,
                                            auid,
                                            group,
                                            gid,
                                            effective_user,
                                            euid,
                                            effective_group,
                                            egid,
                                            cwd,
                                            name
                                            | extend TimeCreatedUtc=TimeGenerated '''}
    _ACCOUNT_LOGON_LX = {'table': 'LinuxAuditD | where EventID == 1100 or EventID == 1112',
                         'query_project': '''| project
                                            TenantId,
                                            Account,
                                            EventID,
                                            TimeGenerated,
                                            SourceComputerId
                                            Computer=node,
                                            SubjectUserName=acct,
                                            SubjectDomainName='',
                                            SubjectUserSid=auid,
                                            TargetUserName=user,
                                            TargetDomainName='',
                                            TargetUserSid=uid,
                                            TargetLogonId=ses,
                                            LogonProcessName=exe,
                                            LogonType=0,
                                            AuthenticationPackageName,
                                            Status=res,
                                            audit_user,
                                            IpAddress=addr,
                                            WorkstationName=hostname'''}
    # Add to the main dictionaries
    DATA_MAPPINGS[DataEnvironment.LogAnalytics][DataFamily.SecurityAlert] = {
        'security_alert': _SECURITY_ALERT}
    DATA_MAPPINGS[DataEnvironment.LogAnalytics][DataFamily.WindowsSecurity] = {
        'process_create': _PROC_CREATE_WIN, 'account_logon': _ACCOUNT_LOGON_WIN,
        'account_logon_fail': _ACCOUNT_LOGON_FAIL_WIN}
    DATA_MAPPINGS[DataEnvironment.LogAnalytics][DataFamily.LinuxSecurity] = {
        'process_create': _PROC_CREATE_LX, 'account_logon': _ACCOUNT_LOGON_LX}

    def __init__(self, environment: DataEnvironment = DataEnvironment.LogAnalytics,
                 data_family: DataFamily = DataFamily.WindowsSecurity,
                 data_source: str = 'security_alert'):
        """
        Create a new instance of the DataSchema object.

            :param environment='LogAnalytics': Specify the environment for the schema.
            :param data_family=DataFamily.WindowsSecurity: Specify the data family
                for the schema (e.g. DataFamily.SecurityAlert)
            :param data_source='security_alert': Specify the data source required.
        """
        if isinstance(environment, str):
            try:
                environment = DataEnvironment[environment]
            except KeyError:
                pass
        if environment not in DataSchema.DATA_MAPPINGS:
            raise LookupError(f'Unknown environment {environment}. '
                              'Valid environments are:\n{self.environments}')

        if isinstance(data_family, str):
            try:
                data_family = DataFamily[data_family]
            except KeyError:
                pass
        if data_family not in DataSchema.DATA_MAPPINGS[environment]:
            raise LookupError(f'Unknown data_family {data_family}. '
                              'Valid families are:\n{self.data_families}')

        if data_source not in DataSchema.DATA_MAPPINGS[environment][data_family]:
            raise LookupError(f'Unknown data_source {data_family}. '
                              'Valid data sources are:\n{self.data_source_types}')

        self.current = DataSchema.DATA_MAPPINGS[environment][data_family][data_source]

    def __getitem__(self, key):
        """Index operator overload."""
        if self.current is not None and key is not None and key in self.current:
            return self.current[key]
        raise KeyError('{} has no attribute {}'.format(str(type(self)), key))

    def __contains__(self, key):
        """In operator overload."""
        if self.current is not None and key is not None and key in self.current:
            return True
        return False

    @property
    def data_environments(self):
        """Return the environments defined in the schema."""
        return self.get_data_environments()

    @property
    def data_families(self):
        """Return the data families in the schema."""
        return self.get_data_families()

    @property
    def data_source_types(self):
        """Return the data families in the schema."""
        sources = set()
        for _, data in self.DATA_MAPPINGS.items():
            for _, data_source in data.items():
                for source in data_source:
                    sources.add(source)

        return sources

    @property
    def property_names(self):
        """Return the current items available in the schema."""
        return list(self.current.keys())

    @classmethod
    def default_schemas(cls, environment=None, data_family=None):
        """
        Return data sources for specified environment and data family.

            :param environment: the data environment
            :param data_family: the data family (e.g. WindowsSecurity,
                LinuxSecurity, Office365)
        """
        if environment is None or environment not in DataSchema.DATA_MAPPINGS:
            raise LookupError('Invalid value for environment. Expected one of {}'
                              .format(', '.join(cls.get_data_environments())))

        env = DataSchema.DATA_MAPPINGS[environment]
        if data_family is None or data_family not in env:
            raise LookupError('Invalid value for data_family. Expected one of {}'
                              .format(', '.join(cls.get_data_families())))

        return env[data_family]

    @classmethod
    def get_data_environments(cls):
        """Return the environments defined in the schema."""
        return list({DataEnvironment(env_name).name for env_name in cls.DATA_MAPPINGS})

    @classmethod
    def get_data_families(cls):
        """Return the data families in the schema."""
        families = set()
        for _, data in cls.DATA_MAPPINGS.items():
            for data_map in data:
                families.add(DataFamily(data_map).name)

        return list(families)


if __name__ == '__main__':
    # mini test code
    # pylint: disable=locally-disabled, C0103
    schema = DataSchema(environment='LogAnalytics',
                        data_family='WindowsSecurity', data_source='process_create')
    assert len(schema.data_environments) >= 2
    assert len(schema.data_families) >= 2
    assert len(schema.data_source_types) >= 3
    assert 'table' in schema
    assert 'query_project' in schema
    schemas = DataSchema.default_schemas(
        environment='LogAnalytics', data_family='WindowsSecurity')
    assert schemas is not None
    assert len(schemas) >= 3
    # pylint: enable=locally-disabled, C0103



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""Module for SecurityAlert class."""
import html
import re
from datetime import datetime
from collections import Counter

import pandas as pd

from .entityschema import Process
from .query_defns import QueryParamProvider, DataFamily, DataEnvironment
from .utility import is_not_empty, escape_windows_path
from . utility import export
from .. _version import VERSION

__version__ = VERSION
__author__ = 'Ian Hellen'

_ID_PROPERTIES = ['AzSubscriptionId', 'AzResourceId', 'WorkspaceId', 'AgentId',
                  'TenantId', 'SourceComputerId', 'ResourceId',
                  'WorkspaceSubscriptionId', 'WorkspaceResourceGroup',
                  'ProviderAlertId', 'SystemAlertId', 'ResourceId']


@export
class SecurityBase(QueryParamProvider):
    """
    Security Base Class for alerts and events.

    Instantiates a security event or alert from a pandas Series
    and provides convenience access methods to retrieve properties.
    """

    def __init__(self, src_row: pd.Series = None):
        """Instantiate a security alert from a pandas Series."""
        self._source_data = src_row
        self._custom_query_params = {}
        self._entities = []

        # Extract and cache alert ID properties
        self._ids = dict()
        for id_property in _ID_PROPERTIES:
            if id_property in self._source_data:
                self._ids[id_property] = self._source_data[id_property]

        self.path_separator = '\\'
        self.os_family = 'Windows'

    # Dunder methods
    def __getitem__(self, key):
        """Allow property get using dictionary key syntax."""
        if key in self.__dict__:
            return self.__dict__[key]
        elif key in self._source_data:
            return self._source_data[key]
        else:
            raise KeyError

    def __setitem__(self, key, value):
        """Allow property set using dictionary key syntax."""
        self.__dict__[key] = value

    def __contains__(self, key):
        """Allow property 'in' test."""
        # get attribute from source dictionary if not defined
        return key in self.__dict__ or key in self._source_data

    def __getattr__(self, name):
        """Return the value of the named property 'name'."""
        if name in self._source_data:
            return self._source_data[name]
        raise AttributeError(f'{name} is not a valid attribute.')

    def __str__(self):
        """Return string representation of object properties."""
        str_props = [f'{prop}: {val}' for prop, val in
                     self._source_data.items()]

        if self.entities:
            str_entities = []
            for ent in self.entities:
                str_entities.append(str(ent).replace('\n', ', '))
            str_props = str_props + str_entities
        return '\n'.join(str_props)

    # def __getstate__(self):
    #     """Return dictionary of state for serialization/pickling."""
    #     state_dict = {}
    #     state_dict['source_data'] = (self._source_data)
    #     state_dict['Entities'] = self._entities
    #     state_dict['path_separator'] = self.path_separator
    #     state_dict['os_family'] = self.os_family
    #     return state_dict

    # def __setstate__(self, state):
    #     """Set state from dictionary for deserialization/unpickling."""
    #     self._source_data = state['source_data']
    #     self._entities = list(state['Entities'])
    #     self.path_separator = state['path_separator']
    #     self.os_family = state['os_family']

    # Properties
    @property
    def entities(self):
        """Return a list of the Security Alert entities."""
        return self._entities

    @property
    def hostname(self):
        """Return the Hostname (not FQDN) of the host associated with the alert."""
        return self.get_entity_property(entity_type='host', entity_property='HostName')

    @property
    def computer(self):
        """
        Return the Computer name of the host associated with the alert.

        (host FQDN, if available)
        """
        return self.primary_host.computer if self.primary_host is not None else None

    @property
    def ids(self):
        """Return a collection of Identity properties for the alert."""
        return self._ids

    @property
    def is_in_workspace(self):
        """Return True if the alert has a Log Analytics WorkspaceID."""
        return 'WorkspaceId' in self._ids and 'AgentId' in self._ids

    @property
    def is_in_log_analytics(self):
        """Return True if the alert originates from a Log Analytics Workspace host."""
        return 'TenantId' in self._ids

    @property
    def is_in_azure_sub(self):
        """Return True if the alert originates from an Azure Security Center host."""
        if ('AzSubscriptionId' not in self._ids and
                'AzResourceId' not in self._ids and
                'ResourceId' in self._ids and self._ids['ResourceId']):
            self._ids['AzResourceId'] = self._id['ResourceId']
            self._ids['AzSubscriptionId'] = (
                self._get_subscription_from_resource(self._id['ResourceId']))

        return 'AzSubscriptionId' in self._ids and 'AzResourceId' in self._ids

    @property
    def primary_host(self):
        """Return the Primary host entity of the host associated with the alert."""
        hosts = self.get_entities_of_type('host')
        if hosts:
            return hosts[0]
        return None

    @property
    def primary_process(self):
        """Return the primary process entity (if any) associated with the alert."""
        procs = self.get_entities_of_type('process')
        if not procs:
            return None
        if len(procs) == 1:
            return procs[0]

        # find the first process that has a parent process property
        procs_with_parent = [p for p in procs if 'ParentProcess' in p]
        return procs_with_parent[0] if procs_with_parent else procs[0]

    @property
    def primary_account(self):
        """Return the primary account entity (if any) associated with this object."""
        accts = self.get_entities_of_type('account')
        return accts[0] if accts else None

    @property
    def query_params(self):
        """
        Query parameters derived from alert.

        Returns:
            dict(str, str) -- Dictionary of parameter names

        """
        try:
            if self.primary_host:
                host_name = self.primary_host.fqdn
            else:
                host_name = None
            proc_name = (self.primary_process.ImageFile.FullPath if
                         self.primary_process and self.primary_process.ImageFile
                         else None)
            acct_name = self.primary_account.Name if self.primary_account else None
            path_separator = self.path_separator
            if self.data_family == DataFamily.WindowsSecurity:
                proc_name = escape_windows_path(proc_name)
                path_separator = escape_windows_path(self.path_separator)

            dyn_query_params = {
                'subscription_filter': self.subscription_filter(),
                'host_filter_eq': self.host_filter(operator='=='),
                'host_filter_neq': self.host_filter(operator='!='),
                'host_name': host_name,
                'account_name': acct_name,
                'process_name': proc_name,
                'logon_session_id': self.get_logon_id(),
                'process_id': (self.primary_process.ProcessId if self.primary_process else None),
                'path_separator': path_separator,
                'data_family': self.data_family,
                'data_environment': self.data_environment,
            }

            dyn_query_params.update(self._custom_query_params)
            return dyn_query_params
        except AttributeError:
            return {}

    @property
    def data_family(self) -> DataFamily:
        """Return the data family of the alert for subsequent queries."""
        if self.os_family == 'Linux':
            return DataFamily.LinuxSecurity
        elif self.os_family == 'Windows':
            return DataFamily.WindowsSecurity
        return None

    @property
    def data_environment(self) -> DataEnvironment:
        """Return the data environment of the alert for subsequent queries."""
        if self.is_in_log_analytics:
            return DataEnvironment.LogAnalytics
        else:
            return DataEnvironment.Kusto

    @property
    def origin_time(self) -> datetime:
        """Return the datetime of event."""
        return self.TimeGenerated

    def get_entity_property(self, entity_property, entity_type=None, entity=None):
        """
        Return the value of the named entity property.

            :param entity_property: the name of the property to return
            :param entity_type=None: the name of the entity type (optional if entity is supplied)
            :param entity=None: the target entity.

        If the entity parameter is not supplied the function will return the
        property value of the first entity of the current alert that
        matches the specified type and has a property of entity_property

        """
        if entity and entity_property in entity:
            return entity[entity_property]

        if self.entities is not None:
            for test_entity in [entity for entity in self.entities
                                if entity['Type'] == entity_type]:
                if (test_entity and entity_property in test_entity and
                        is_not_empty(test_entity[entity_property])):
                    return test_entity[entity_property]
        return None

    def get_logon_id(self, account=None):
        """
        Get the logon Id for the alert or the account, if supplied.

        if the account entity is not supplied, return the logon id
        of the first host-logon-session or account entity.
        """
        for session in [e for e in self.entities if
                        e['Type'] == 'host-logon-session' or e['Type'] == 'hostlogonsession']:
            if account is None or session['Account'] == account:
                return session['SessionId']
        if account is None:
            for acct in [e for e in self.entities if e['Type'] == 'account' and 'LogonId' in e]:
                return acct['LogonId']
        elif 'LogonId' in account:
            return account['LogonId']
        return None

    def get_process_name(self, process):
        """Return the process (filename) of the process."""
        if isinstance(process, Process) and process.ProcessFilePath:
            return process.ProcessFilePath
        if 'ImageFile' in process:
            if 'FullPath' in process['ImageFile']:
                return process['ImageFile']['FullPath']
            elif 'Directory' in process['ImageFile']:
                return (process['ImageFile']['Directory'] +
                        self.path_separator + process['ImageFile']['Name'])
        return None

    def subscription_filter(self, operator='=='):
        """Return a KQL subscription filter clause derived from the alert properties."""
        if self.is_in_log_analytics:
            return 'true'
        if self.is_in_azure_sub:
            return ('AzureResourceSubscriptionId {} \'{}\''
                    .format(operator, self._ids['AzSubscriptionId']))
        if self.is_in_workspace:
            return 'WorkspaceId {} \'{}\''.format(operator, self._ids['WorkspaceId'])
        # Otherwise we default to including everything
        return 'true'

    def host_filter(self, operator='=='):
        """
        Return a KQL host filter clause derived from the alert properties.

            :param operator='==': the operator to use in the filter clause.
                '==' and '!=' typically.
        """
        if (self.is_in_log_analytics and 'SourceComputerId' in self._ids and
                self._ids['SourceComputerId']):
            return 'SourceComputerId {} \'{}\''.format(operator, self._ids['SourceComputerId'])
        if (self.is_in_azure_sub and 'AzureResourceId' in self._ids and
                self._ids['AzResourceId']):
            return 'AzureResourceId {} \'{}\''.format(operator, self._ids['AzResourceId'])
        if self.is_in_workspace and 'AgendId' in self._ids and self._ids['AgentId']:
            return 'AgentId {} \'{}\''.format(operator, self._ids['AgentId'])

        if self.primary_host:
            case_insens_op = '=~' if operator == '==' else '!~'
            return 'Computer {} \'{}\''.format(case_insens_op, self.primary_host.computer)
        return None

    def get_entities_of_type(self, entity_type='host'):
        """Return entity collection for a give entity type."""
        return [p for p in self.entities if p['Type'] == entity_type]

    def to_html(self, show_entities=False):
        """Return the item as HTML string."""
        html_doc = pd.DataFrame(self._source_data).to_html()

        if 'ExtendedProperties' in self._source_data:
            ext_prop_title = '<br/><h3>ExtendedProperties:</h3>'
            ext_prop_html = pd.DataFrame(
                pd.Series(self._source_data['ExtendedProperties'])).to_html()
            html_doc = html_doc + ext_prop_title + ext_prop_html

        if show_entities and self.entities:
            entity_title = '<br/><h3>Entities:</h3><br/>'
            entity_html = '<br/>'.join([self._format_entity(ent)
                                        for ent in self.entities])
            html_doc = html_doc + entity_title + entity_html
        else:
            e_counts = Counter([ent['Type'] for ent in self.entities])
            e_counts_str = ', '.join(
                [f'{e}: {c}' for e, c in e_counts.items()])
            html_doc = html_doc + f'<h3>Entity counts: </h3>{e_counts_str}'
        return html_doc

    def _format_entity(self, entity):
        str_entity = str(entity)
        if str_entity:
            str_entity = str_entity.replace('\n', ', ')
        return html.escape(str_entity)

    def _find_os_family(self):
        """Work out which OSFamily and path separator to use from entities or file paths."""
        # Use OSFamily if any entities have this property set
        os_family_entities = [e for e in self.entities if 'OSFamily' in e]
        if os_family_entities:
            for os_entity in os_family_entities:
                if os_entity['OSFamily'] == 'Linux':
                    self.os_family = 'Linux'
                    self.path_separator = '/'
                    break
        else:
            # Otherwise try to infer from the file paths
            files = [e for e in self.entities if e['Type'] == 'file']
            if files:
                for file in files:
                    if 'Directory' in file and '/' in file['Directory']:
                        self.os_family = 'Linux'
                        self.path_separator = '/'
                        break
            else:
                for proc in [e for e in self.entities
                             if e['Type'] == 'process' and 'ImageFile' in e]:
                    file = proc['ImageFile']
                    if 'Directory' in file and '/' in file['Directory']:
                        self.os_family = 'Linux'
                        self.path_separator = '/'
                        break

    @staticmethod
    def _get_subscription_from_resource(resource_id):
        """Extract subscription Id from resource string."""
        sub_regex = r'^/subscriptions/([^/]+)/'
        sub_ids = re.findall(sub_regex, resource_id, re.RegexFlag.I)
        if sub_ids:
            return sub_ids[0]




import unittest

import pandas as pd

# Test code
from .. msticpy.sectools.iocextract import IoCExtract

TEST_CASES = {
    'ipv4_test': r'c:\one\path\or\another\myprocess -ip4:206.123.1.123',
    'ipv6_test': r'''c:\one\path\or\another\myprocess -ip6:(2001:0db8:85a3:0000:0000:8a2e:0370:7334,
    2001:db8:85a3:0:0:8a2e:370:7334,2001:db8:85a3::8a2e:370:7334,::ffff:192.0.2.128)''',
    'url_test': r'''c:\one\path\or\another\myprocess /url:https://some.domain.it/thepath?qry1=abc&qry2=xyz
        /url:https://myuser@some.domain.es:88/thepath?qry1=abc&qry2=xyz''',
    'windows_path_test': r'c:\one\path\or\another\myprocess -file:"..\another\file" -file:"\\uncpath\file"',
    'linux_path_test': r'/bin/bash --file:./bish --file:/bin/bash --file:../../bosh',
    'md5_hash_test': '00236a2ae558018ed13b5222ef1bd987hash -something-hash=00236a2ae558018ed13b5222ef1bd987hash -something -hash=00236a2ae558018ed13b5222ef1bd987',
    'sha1_hash_test': '00236a2ae558018ed13b5222ef1bd98700000001hash -something -hash=00236a2ae558018ed13b5222ef1bd98700000001hash -something -hash=00236a2ae558018ed13b5222ef1bd98700000001',
    'sha256_hash_test': '''00236a2ae558018ed13b5222ef1bd98700000001123456789012345678901234hash -something -hash=00236a2ae558018ed13b5222ef1bd98700000001123456789012345678901234hash -something
-hash=00236a2ae558018ed13b5222ef1bd98700000001123456789012345678901234''',
    'url2_test': 'curl \'https://www.virustotal.com/en/ip-address/90.156.201.27/information/\''}


class TestIoCExtractor(unittest.TestCase):
    """Unit test class."""
    def __run_extract(self, extractor=None, testcase=None, expected_items=None,
                      os_family='Windows'):
        if extractor is None or testcase is None or expected_items is None:
            raise Exception('One or more required parameters were missing')

        test_input = TEST_CASES[testcase + '_test']
        results = extractor.extract(test_input, os_family=os_family)
        for k, v in expected_items.items():
            self.assertEqual(len(results[k]), v, 'Unexpected value for ' + k)

    def setUp(self):
        self.extractor = IoCExtract()

    def test_ipv4(self):
        self.__run_extract(self.extractor, 'ipv4', {'ipv4': 1})

    def test_ipv6(self):
        self.__run_extract(self.extractor, 'ipv6', {'ipv6': 2})

    def test_url(self):
        self.__run_extract(self.extractor, 'url', {'url': 2, 'dns': 2, 'ipv4': 0})
        self.__run_extract(self.extractor, 'url2', {'url': 1, 'dns': 1, 'ipv4': 1})

    def test_windows_path(self):
        self.__run_extract(self.extractor, 'windows_path', {'windows_path': 3})

    def test_linux_path(self):
        self.__run_extract(self.extractor, 'linux_path', {'linux_path': 3},
                           os_family='Linux')

    def test_hashes(self):
        self.__run_extract(self.extractor, 'md5_hash', {'md5_hash': 3})
        self.__run_extract(self.extractor, 'sha1_hash', {'sha1_hash': 3})
        self.__run_extract(self.extractor, 'sha256_hash', {'sha256_hash': 3})

    def test_dataframe(self):

        input_df = pd.DataFrame.from_dict(data=TEST_CASES, orient='index', columns=['input'])
        output_df = self.extractor.extract(data=input_df, columns=['input'], os_family='Windows')

        self.assertGreater(output_df.shape[0], 0)
        self.assertEqual(output_df[output_df['IoCType'] == 'ipv4'].shape[0], 3)
        self.assertEqual(output_df[output_df['IoCType'] == 'ipv6'].shape[0], 2)
        self.assertEqual(output_df[output_df['IoCType'] == 'url'].shape[0], 3)
        self.assertEqual(output_df[output_df['IoCType'] == 'windows_path'].shape[0], 6)
        self.assertEqual(output_df[output_df['IoCType'] == 'linux_path'].shape[0], 0)
        self.assertEqual(output_df[output_df['IoCType'] == 'md5_hash'].shape[0], 3)
        self.assertEqual(output_df[output_df['IoCType'] == 'sha1_hash'].shape[0], 3)
        self.assertEqual(output_df[output_df['IoCType'] == 'sha256_hash'].shape[0], 3)

        input_df = pd.DataFrame.from_dict(data=TEST_CASES, orient='index', columns=['input'])
        output_df = self.extractor.extract(data=input_df, columns=['input'], os_family='Linux')
        # for _, row in output_df[output_df['IoCType'] == 'url'].iterrows():
        #     print(row.Observable)
        self.assertGreater(output_df.shape[0], 0)
        self.assertEqual(output_df[output_df['IoCType'] == 'ipv4'].shape[0], 3)
        self.assertEqual(output_df[output_df['IoCType'] == 'ipv6'].shape[0], 2)
        self.assertEqual(output_df[output_df['IoCType'] == 'url'].shape[0], 3)
        self.assertEqual(output_df[output_df['IoCType'] == 'windows_path'].shape[0], 0)
        self.assertEqual(output_df[output_df['IoCType'] == 'linux_path'].shape[0], 8)
        self.assertEqual(output_df[output_df['IoCType'] == 'md5_hash'].shape[0], 3)
        self.assertEqual(output_df[output_df['IoCType'] == 'sha1_hash'].shape[0], 3)
        self.assertEqual(output_df[output_df['IoCType'] == 'sha256_hash'].shape[0], 3)

    def test_dataframe_ioc_types(self):

        input_df = pd.DataFrame.from_dict(data=TEST_CASES, orient='index', columns=['input'])
        output_df = self.extractor.extract(data=input_df, columns=['input'], os_family='Windows',
            ioc_types=['ipv4', 'url', 'md5_hash'])

        self.assertGreater(output_df.shape[0], 0)
        self.assertEqual(output_df[output_df['IoCType'] == 'ipv4'].shape[0], 3)
        self.assertEqual(output_df[output_df['IoCType'] == 'ipv6'].shape[0], 0)
        self.assertEqual(output_df[output_df['IoCType'] == 'url'].shape[0], 3)
        self.assertEqual(output_df[output_df['IoCType'] == 'windows_path'].shape[0], 0)
        self.assertEqual(output_df[output_df['IoCType'] == 'linux_path'].shape[0], 0)
        self.assertEqual(output_df[output_df['IoCType'] == 'md5_hash'].shape[0], 3)
        self.assertEqual(output_df[output_df['IoCType'] == 'sha1_hash'].shape[0], 0)
        self.assertEqual(output_df[output_df['IoCType'] == 'sha256_hash'].shape[0], 0)

if __name__ == '__main__':
    unittest.main()







# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""geoip module using ipstack."""
# import gzip
from json import JSONDecodeError
import math
import os
from abc import ABC, abstractmethod
from collections.abc import Iterable
from datetime import datetime, timedelta
from typing import Tuple
from IPython import get_ipython
from IPython.display import display, HTML

from geolite2 import geolite2
import pandas as pd
import requests

from .. nbtools.entityschema import GeoLocation, IpAddress
from .. nbtools.utility import export
from .. _version import VERSION

__version__ = VERSION
__author__ = 'Ian Hellen'


class GeoIpLookup(ABC):
    """Abstract base class for GeoIP Lookup classes."""

    @abstractmethod
    def lookup_ip(self, ip_address: str = None, ip_addr_list: Iterable = None,
                  ip_entity: IpAddress = None):
        """
        Lookup IP location.

        Keyword Arguments:
            ip_address {str} -- a single address to look up (default: {None})
            ip_addr_list {Iterable} -- a collection of addresses to lookup (default: {None})
            ip_entity {IpAddress} -- an IpAddress entity

        Returns:
            tuple(list{dict}, list{entity}) -- returns raw geolocation results and
                same results as IP/Geolocation entities

        """

    def df_lookup_ip(self, data: pd.DataFrame, column: str):
        """
        Lookup Geolocation data from a pandas Dataframe.

        Keyword Arguments:
            data {pd.DataFrame} -- pandas dataframe containing IpAddress column
            column {str} -- the name of the dataframe column to use as a source
        """
        ip_list = data[column].values
        _, entities = self.lookup_ip(ip_addr_list=ip_list)

        ip_dicts = [ent.Location.properties().update(IpAddress=ent.Address)
                    for ent in entities]
        df_out = pd.DataFrame(data=ip_dicts)
        return data.merge(df_out, left_on=column, right_on='IpAddress')


@export
class IPStackLookup(GeoIpLookup):
    """
    GeoIP Lookup using IPStack web service.

    Raises:
        ConnectionError -- Invalid status returned from http request
        PermissionError -- Service refused request (e.g. requesting batch of addresses
                on free tier API key)

    """

    _IPSTACK_API = 'http://api.ipstack.com/{iplist}?access_key={access_key}'

    def __init__(self, api_key):
        """
        Create a new instance of IPStackLookup.

        Arguments:
            api_key {str} -- API Key from IPStack - see https://ipstack.com
        """
        self._api_key = api_key

    def lookup_ip(self, ip_address: str = None, ip_addr_list: Iterable = None,
                  ip_entity: IpAddress = None):
        """
        Lookup IP location from IPStack web service.

        Keyword Arguments:
            ip_address {str} -- a single address to look up (default: {None})
            ip_addr_list {Iterable} -- a collection of addresses to lookup (default: {None})
            ip_entity {IpAddress} -- an IpAddress entity

        Raises:
            ConnectionError -- Invalid status returned from http request
            PermissionError -- Service refused request (e.g. requesting batch of addresses
                on free tier API key)

        Returns:
            tuple(list{dict}, list{entity}) -- returns raw geolocation results and
                same results as IP/Geolocation entities

        """
        if ip_address and isinstance(ip_address, str):
            ip_list = [ip_address.strip()]
        elif ip_addr_list:
            ip_list = (ip.strip() for ip in ip_addr_list)
        elif ip_entity:
            ip_list = [ip_entity.Address]
        else:
            raise ValueError('No valid ip addresses were passed as arguments.')

        results, status = self._submit_request(ip_list)
        if status != 200:
            raise ConnectionError('Error reponse from server: ' + str(status))

        # {"success":false,"error":{"code":303,"type":"batch_not_supported_on_plan",
        # "info":"Bulk requests are not supported on your plan.
        # Please upgrade your subscription."}}

        if 'success' in results and not results["success"]:
            raise PermissionError('Service unable to complete request. Error: {}'
                                  .format(results['error']))

        output_raw = list(results)
        output_entities = []
        if isinstance(results, list):
            for ip_loc in results:
                output_entities.append(self._create_ip_entity(ip_loc, ip_entity))
        else:
            output_entities.append(self._create_ip_entity(results, ip_entity))

        return output_raw, output_entities

    def _create_ip_entity(self, ip_loc: dict, ip_entity):
        if not ip_entity:
            ip_entity = IpAddress()
            ip_entity.Address = ip_loc['ip']
        geo_entity = GeoLocation()
        geo_entity.CountryCode = ip_loc['country_code']

        geo_entity.CountryName = ip_loc['country_name']
        geo_entity.State = ip_loc['region_name']
        geo_entity.City = ip_loc['city']
        geo_entity.Longitude = ip_loc['longitude']
        geo_entity.Latitude = ip_loc['latitude']
        if 'connection' in ip_loc:
            geo_entity.Asn = ip_loc['connection']['asn']
        ip_entity.Location = geo_entity
        return ip_entity

    def _submit_request(self, ip_list):
        """
        Submit the request to IPStack.

            :param ip_list: Comma-separated string list of IPs to look up
        """
        submit_url = self._IPSTACK_API.format(iplist=ip_list, access_key=self._api_key)
        response = requests.get(submit_url)

        if response.status_code == 200:
            return response.json(), response.status_code
        else:
            if response:
                try:
                    return response.json(), response.status_code
                except JSONDecodeError:
                    pass
            return None, response.status_code


@export
class GeoLiteLookup(GeoIpLookup):
    """
    GeoIP Lookup using MaxMindDB database.

    Raises:
        ConnectionError -- Invalid status returned from http request
        PermissionError -- Service refused request (e.g. requesting batch of addresses
                on free tier API key)

    """

    _MAXMIND_DOWNLOAD = 'https://dev.maxmind.com/geoip/geoip2/geolite2/#Downloads'
    _DB_FILE = 'GeoLite2-City.mmdb'

    def __init__(self):
        """Return new instance of GeoLiteLookup class."""
        self._reader = geolite2.reader()
        last_mod_time = datetime.fromtimestamp(os.path.getmtime(geolite2.filename))
        db_age = datetime.utcnow() - last_mod_time
        if db_age > timedelta(40):
            print(f'{self._DB_FILE} is over one month old. Update the maxminddb package')
            print(f'to refresh or download a new version from {self._MAXMIND_DOWNLOAD}')

    def lookup_ip(self, ip_address: str = None, ip_addr_list: Iterable = None,
                  ip_entity: IpAddress = None):
        """
        Lookup IP location from IPStack web service.

        Keyword Arguments:
            ip_address {str} -- a single address to look up (default: {None})
            ip_addr_list {Iterable} -- a collection of addresses to lookup (default: {None})

        Returns:
            tuple(list{dict}, list{entity}) -- returns raw geolocation results and
                same results as IP/Geolocation entities

        """
        if ip_address and isinstance(ip_address, str):
            ip_list = [ip_address.strip()]
        elif ip_addr_list:
            ip_list = (ip.strip() for ip in ip_addr_list)
        elif ip_entity:
            ip_list = [ip_entity.Address]
        else:
            raise ValueError('No valid ip addresses were passed as arguments.')

        output_raw = []
        output_entities = []
        for ip_input in ip_list:
            geo_match = self._reader.get(ip_input)
            if geo_match:
                output_raw.append(geo_match)
                output_entities.append(self._create_ip_entity(ip_input, geo_match, ip_entity))

        return output_raw, output_entities

    def _create_ip_entity(self, ip_address, geo_match: dict, ip_entity):
        if not ip_entity:
            ip_entity = IpAddress()
            ip_entity.Address = ip_address
        geo_entity = GeoLocation()
        geo_entity.CountryCode = geo_match.get('country', {}).get('iso_code', None)
        geo_entity.CountryName = geo_match.get('country', {}).get('names', {}).get('en', None)
        subdivs = geo_match.get('subdivisions', [])
        if subdivs:
            geo_entity.State = subdivs[0].get('names', {}).get('en', None)
        geo_entity.City = geo_match.get('city', {}).get('names', {}).get('en', None)
        geo_entity.Longitude = geo_match.get('location', {}).get('longitude', None)
        geo_entity.Latitude = geo_match.get('location', {}).get('latitude', None)
        ip_entity.Location = geo_entity
        return ip_entity


_MM_LICENSE_HTML = '''
This product includes GeoLite2 data created by MaxMind, available from
<a href="https://www.maxmind.com">https://www.maxmind.com</a>.
'''
_MM_LICENSE_TXT = '''
This product includes GeoLite2 data created by MaxMind, available from
https://www.maxmind.com.
'''
_IPSTACK_LICENSE_HTML = '''
This library uses services provided by ipstack.
<a href="https://ipstack.com">https://ipstack.com</a>'''

_IPSTACK_LICENSE_TXT = 'This library uses services provided by ipstack (https://ipstack.com)'

if not get_ipython():
    print(_MM_LICENSE_TXT)
    print(_IPSTACK_LICENSE_TXT)
else:
    display(HTML(_MM_LICENSE_HTML))
    display(HTML(_IPSTACK_LICENSE_HTML))


def geo_distance(origin: Tuple[float, float],
                 destination: Tuple[float, float]) -> float:
    """
    Calculate the Haversine distance.

    Author: Martin Thoma - stackoverflow

    Parameters
    ----------
    origin : tuple of float
        (lat, long)
    destination : tuple of float
        (lat, long)

    Returns
    -------
    distance_in_km : float

    Examples
    --------
    >>> origin = (48.1372, 11.5756)  # Munich
    >>> destination = (52.5186, 13.4083)  # Berlin
    >>> round(distance(origin, destination), 1)
    504.2

    """
    orig_lat, orig_lon = origin
    dest_lat, dest_lon = destination
    EARTH_RADIUS_KM = 6371  # km

    ang_dist_lat = math.radians(dest_lat - orig_lat)
    ang_dist_lon = math.radians(dest_lon - orig_lon)
    hav_a = (math.sin(ang_dist_lat / 2) * math.sin(ang_dist_lat / 2) +
             math.cos(math.radians(orig_lat)) * math.cos(math.radians(dest_lat)) *
             math.sin(ang_dist_lon / 2) * math.sin(ang_dist_lon / 2))
    hav_c = 2 * math.atan2(math.sqrt(hav_a), math.sqrt(1 - hav_a))
    return EARTH_RADIUS_KM * hav_c



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""
base64_unpack.

The main function of this module is to decode and unpack
strings that are obfuscated using base64 and/or certain
compression algorithms such as gzip and zip.

It has the following functions:
unpack_items - this is the main entry point and takes either a string or
a pandas dataframe (with specified column) as input. It returns a string
with obfuscated parts replaced by decoded equivalents (unless the decoding
results in an undecodable binary, in which case a placeholder is used).


Other helper functions may also be useful standalone
get_items_from_gzip(binary): Return decompressed gzip content of byte string
get_items_from_zip(binary): Return dictionary of zip contents from byte string
get_items_from_tar(binary): Return dictionary of tar file contents
get_hashes(binary): Return md5, sha1 and sha256 hashes of input byte string
"""

import base64
import binascii
import gzip
import hashlib
import io
import re
import tarfile
import zipfile
from collections import namedtuple

import pandas as pd

from .. nbtools.utility import pd_version_23, export
from .. _version import VERSION

__version__ = VERSION
__author__ = 'Ian Hellen'

_RESULT_FIELDS = ['reference', 'original_string', 'file_name', 'file_type',
                  'input_bytes', 'decoded_string', 'encoding_type',
                  'file_hashes', 'md5', 'sha1', 'sha256', 'printable_bytes']

BinaryRecord = namedtuple('BinaryRecord', _RESULT_FIELDS)

# pylint: disable=locally-disabled, line-too-long
_BASE64_HEADER_TYPES = {
    'TVqQAAMAAAAEAAAA//8AALgAAAAAAAAAQAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA+AAAAA4fug': 'exe',
    'TVqQAAMAAAAEAAAA//8AALgAAAAAAAAAQAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA8AAAAA4fug': 'dll',
    'TVqQAAMAAAAEAAAA//8AALgAAAAAAAAAQAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA6AAAAA4fug': 'sys',
    'UEsDBBQAAAAIA': 'zip',
    'UEsDBBQAAQAIA': 'zip (passprotected)',
    'H4sI': 'gz',
    'N3q8ryccAAR': '7z',
    'UmFyIRoHAM': 'rar',
    'JVBERi0xLjcNC': 'pdf',
    '0M8R4KGxGuE': 'msi',
    'TVNXSU0AAADQ': 'wim'}
# pylint: enable=locally-disabled, line-too-long
_BASE64_HEADER_OFFSET_TYPES = {
    'DAxMDA3NzcAMDAwMDAwM': 'tar'
}

# Base64 simple regex
_BASE64_REGEX = '(?P<b64>[A-Za-z0-9+/\\n\\r]{30,}={0,2})'
BASE64_REGEX_C = re.compile(_BASE64_REGEX, re.I | re.X)
# '''(?<![A-Za-z0-9+\/])
#             (?:
#                 (?:[A-Za-z0-9+\/]{4})*
#                 (?:
#                     [A-Za-z0-9+\/]{2} [AEIMQUYcgkosw048] = |
#                     [A-Za-z0-9+\/] [AQgw] ==
#                 ) |
#                 (?:[A-Za-z0-9\/]{4})*
#                 (?:
#                     \+ [A-Za-z0-9\/+]{3} |
#                     [A-Za-z0-9\/] \+ [A-Za-z0-9+\/]{2} |
#                     [A-Za-z0-9\/]{2} \+ [A-Za-z0-9+\/] |
#                     [A-Za-z0-9\/]{3} \+
#                 )
#                 (?:[A-Za-z0-9+\/]{4})*
#             )
#             (?![A-Za-z0-9+\/=])'''

# When True prints see more verbose execution
# (set from 'trace' parameter to unpack_items)
# pylint: disable=locally-disabled, C0103
_debug_trace = False
# pylint: enable=locally-disabled, C0103


@export
def unpack_items(input_string=None, data=None, column=None, trace=False):
    """
    Base64 decode an input string or multiple strings taken from a pandas dataframe column.

    If the input is a dataframe you must supply the name of the column to use.

    Items that decode to utf-8 or utf-16 strings will be returned as decoded
    strings replaced in the original string. If the encoded string is a
    known binary type it will identify the file type and return the hashes
    of the file. If any binary types are known archives (zip, tar, gzip) it
    will unpack the contents of the archive.
    For any binary it will return the decoded file as a byte array, and as a
    printable list of byte values. If the input is a string the function
    returns:

    - decoded string: this is the input string with any decoded sections
    replaced by the results of the decoding

    It also returns the data as a Pandas DataFrame with the following columns:
    - reference : this is an index that matches an index number in the
    returned string (e.g. <<encoded binary type=pdf index=1.2').
    - original_string : the string prior to decoding - file_type : the type
    of file if this could be determined
    - file_hashes : a dictionary of hashes (the md5, sha1 and sha256 hashes
    are broken out into separate columns)
    - input_bytes : the binary image as a byte array
    - decoded_string : printable form of the decoded string (either string
    or list of hex byte values)
    - encoding_type : utf-8, utf-16 or binary
    - md5, sha1, sha256 : the respective hashes of the binary file_type,
    file_hashes, input_bytes, md5, sha1, sha256 will be null if this item is
    decoded to a string

    If the input is a dataframe the output dataframe will also include the
    following column: - src_index - the index of the source row in the input
    frame. This allows you to re-join the output data to the input data
    """
    global _debug_trace
    _debug_trace = trace

    if input_string is not None:
        return _decode_b64_string_recursive(input_string)
    if data is not None:
        if not column:
            raise ValueError('column must be supplied if the input is a DataFrame')

        output_df = None
        for input_row in data[[column]].itertuples():
            (decoded_string, output_frame) = _decode_b64_string_recursive(input_row[1])
            output_frame['src_index'] = input_row.Index
            output_frame['full_decoded_string'] = decoded_string
            if output_df is None:
                output_df = output_frame
            else:
                if pd_version_23():
                    output_df = output_df.append(output_frame, ignore_index=True,
                                                 sort=False)
                else:
                    output_df = output_df.append(output_frame, ignore_index=True)
        return output_df
    return None


def _debug_print_trace(*args):
    if _debug_trace:
        for arg in args:
            print(arg, end='')
        print()


# base64 decoding
def _decode_and_format_b64_string(b64encoded_string, item_prefix=None,
                                  current_depth=1, current_index=1):
    """Decode string and return displayable content plus list of decoded artifacts."""
    # Check if we recognize this as a known file type
    (_, f_type) = _is_known_b64_prefix(b64encoded_string)
    _debug_print_trace('Found type: ', f_type)
    output_files = _decode_b64_binary(b64encoded_string, f_type)
    if not output_files:
        return b64encoded_string, None

    if len(output_files) == 1:
        # get the first (only) item
        out_name, out_record = list(output_files.items())[0]
        _debug_print_trace('_decode_b64_binary returned a single record')
        _debug_print_trace('record:', out_record)

        # Build display string
        # If a string, include the decoded item in the output
        if out_record.encoding_type in ['utf-8', 'utf-16']:
            display_string = f'<decoded type=\'string\' name=\'{out_name}\' ' +\
                             f'index=\'{item_prefix}{current_index}\' ' +\
                             f'depth=\'{current_depth}\'>' +\
                             f'{out_record.decoded_string}</decoded>'
            return display_string, [out_record]
        else:
            # if a binary just record its presence
            display_string = f'<decoded value=\'binary\'  name=\'{out_name}\' ' +\
                             f'type=\'{out_record.file_type}\' ' +\
                             f'index=\'{item_prefix}{current_index}\' ' +\
                             f'depth=\'{current_depth}\'/>'
            return display_string, [out_record]
    else:
        # Build header display string
        display_header = f'<decoded value=\'multiple binary\' type=\'multiple\' ' +\
                         f' index=\'{item_prefix}{current_index}\'>'
        child_display_strings = []
        child_index = 1
        child_depth = current_depth + 1
        _debug_print_trace('_decode_b64_binary returned multiple records')

        # Build child display strings
        for child_name, child_rec in output_files.items():
            _debug_print_trace('Child_decode: ', child_rec)
            child_index_string = f'{item_prefix}{current_index}.{child_index}'

            if child_rec.encoding_type in ['utf-8', 'utf-16']:
                # If a string, include the decoded item in the output
                child_display_string = f'<decoded type=\'string\' name=\'{child_name}\' ' +\
                                       f'index=\'{child_index_string}\' ' +\
                                       f'depth=\'{child_depth}\'>' +\
                                       f'{child_rec.decoded_string}</decoded>'
            else:
                # if a binary just record its presence
                child_display_string = f'<decoded type=\'{child_rec.file_type}\' ' +\
                                       f'name=\'{child_name}\' ' +\
                                       f'index=\'{child_index_string}\' ' +\
                                       f'depth=\'{child_depth}\'/>'
            child_display_strings.append(child_display_string)
            child_index += 1

        display_string = display_header + ''.join(child_display_strings) + '</decoded>'
        return display_string, output_files.values()


def _decode_b64_string_recursive(input_string, undecodable_strings=None,
                                 max_recursion=20, current_depth=1, item_prefix=''):
    """
    Recursively decode and unpack an encoded string.

        :rtype: (str, [BinaryRecord]
    """
    _debug_print_trace('_decode_b64_string_recursive: ', max_recursion)
    _debug_print_trace('processing input: ', input_string[:200])

    decoded_string = input_string

    # we use this to store a set of strings that match the B64 regex but
    # that we were unable to decode - so that we don't end up in an
    # infinite loop
    if not undecodable_strings:
        undecodable_strings = set()

    binary_records = pd.DataFrame(columns=_RESULT_FIELDS)
    fragment_index = 0
    match_pos = 0
    something_decoded = False
    while True:
        # search sequentially through the input string for any strings that look like base64
        _debug_print_trace('regex searching ', decoded_string[:200],
                           ' from pos: ', match_pos, ' bin_index ',
                           fragment_index)
        b64match = BASE64_REGEX_C.search(decoded_string, match_pos)

        # _debug_print_trace('groups: ', len(b64match.groupdict()))
        if b64match is not None:
            _debug_print_trace('regex found: ', b64match.groupdict()['b64'])
            # if (in a recursive call) we already know that this string doesn't decode
            # skip this match
            if b64match.groupdict()['b64'] in undecodable_strings:
                _debug_print_trace('previous undecodable string')
                match_pos = b64match.end()
                continue

            # try to decode
            fragment_index += 1
            (decoded_fragment, binary_items) = (
                _decode_and_format_b64_string(b64match.groupdict()['b64'],
                                              item_prefix=item_prefix,
                                              current_index=fragment_index,
                                              current_depth=current_depth))

            # if the string didn't decode we'll have the same output as input
            # so add that to our set of undecodable strings (we need to track this
            # otherwise we will recurse infinitely)
            something_decoded = (decoded_fragment != b64match.groupdict()['b64'])
            if something_decoded:
                # we did decode something so lets put our result this in the output string
                if binary_items:
                    # if there are one or more binary_items from decoding
                    # add them to our output DataFrame
                    for bin_record in binary_items:
                        new_row = bin_record._asdict()
                        new_row['reference'] = f'{item_prefix}{current_depth}.{fragment_index}'
                        new_row['original_string'] = b64match.groupdict()['b64']
                        new_row['md5'] = new_row['file_hashes']['md5']
                        new_row['sha1'] = new_row['file_hashes']['sha1']
                        new_row['sha256'] = new_row['file_hashes']['sha256']

                        if pd_version_23():
                            binary_records = binary_records.append(new_row,
                                                                   ignore_index=True,
                                                                   sort=False)
                        else:
                            binary_records = binary_records.append(new_row,
                                                                   ignore_index=True)

                # replace the decoded fragment in our current results string (decode_string)
                decoded_string = decoded_string.replace(b64match.groupdict()['b64'],
                                                        decoded_fragment)
                _debug_print_trace('Replaced string', decoded_string[match_pos:match_pos + 100])
                match_pos += len(decoded_fragment)
            else:
                undecodable_strings.add(b64match.groupdict()['b64'])
                _debug_print_trace('new undecodable string')
                match_pos = b64match.end()

        else:
            _debug_print_trace('Regex not found: ', match_pos)
            break

        if fragment_index > 50:
            break

    # if we reach our max recursion depth bail out here
    if max_recursion == 0:
        _debug_print_trace('max recursion reached')
        return decoded_string, binary_records

    if something_decoded:
        # stuff that we have already decoded may also contain further base64 encoded strings
        pfx = f'{item_prefix}.{fragment_index}.' if item_prefix else f'{fragment_index}.'
        (next_level_string, child_records) = _decode_b64_string_recursive(
            decoded_string, undecodable_strings, item_prefix=pfx,
            max_recursion=max_recursion - 1, current_depth=(current_depth + 1))
        return next_level_string, binary_records.append(
            child_records, ignore_index=True, sort=False)
    else:
        _debug_print_trace('Nothing left to decode')
        return decoded_string, binary_records


def _print_bytes(bytes_array):
    """Print byte array as string or hex."""
    enc_results = _get_byte_encoding(bytes_array)
    if enc_results.encoding_type != 'binary':
        print(enc_results.decoded_string)

    else:
        print('Could not decode bytes to string. Hashes:')
        print(get_hashes(_binary_to_bytesio(bytes_array).getbuffer()))
        print(enc_results.printable_bytes)


def _as_byte_string(bytes_array):
    return ' '.join(['{0:02x}'.format(b) for b in bytes_array])


def _empty_binary_rec():
    return BinaryRecord(reference=None,
                        original_string=None,
                        decoded_string=None,
                        encoding_type=None,
                        file_name=None,
                        file_type=None,
                        input_bytes=None,
                        file_hashes=None,
                        md5=None,
                        sha1=None,
                        sha256=None,
                        printable_bytes=None)


def _get_byte_encoding(bytes_array):
    """
    Return encoding type and decoded result.

    Decoded result is list of hex bytes in the case that the decoded
    result is not a string
    """
    result_rec = _empty_binary_rec()
    printable_bytes = _as_byte_string(bytes_array)
    try:
        decoded_string = bytes_array.decode('utf-8')
        return result_rec._replace(decoded_string=decoded_string,
                                   encoding_type='utf-8',
                                   printable_bytes=printable_bytes)
    except UnicodeDecodeError:
        pass

    try:
        # Difficult to tell the difference between a real unicode string
        # and a binary string that happens to decode to a utf-16 string
        decoded_string = bytes_array.decode('utf-16')
        return result_rec._replace(decoded_string=decoded_string,
                                   encoding_type='utf-16',
                                   printable_bytes=printable_bytes)
    except UnicodeDecodeError:
        pass
    return result_rec._replace(encoding_type='binary',
                               printable_bytes=printable_bytes)


def _is_known_b64_prefix(input_string):
    """If this is known file type return the prefix and file type."""
    first160chars = input_string[0:160].replace('\n', '').replace('\r', '')
    for prefix, file_type in _BASE64_HEADER_TYPES.items():
        if first160chars.startswith(prefix):
            return prefix, file_type

    for matching_string, file_type in _BASE64_HEADER_OFFSET_TYPES.items():
        regex_comp = re.compile(matching_string, re.I | re.X)
        if regex_comp.search(first160chars):
            return matching_string, file_type
    return None, None


def _decode_b64_binary(input_string, file_type=None):
    """Examine input string for known binaries and decode and unpack."""
    if not file_type:
        (_, f_type) = _is_known_b64_prefix(input_string)
        file_type = f_type

    try:
        decoded_bytes = base64.b64decode(input_string)
        return _unpack_and_hash_b64_binary(decoded_bytes, file_type)
    except binascii.Error:
        # we couldn't decode
        _debug_print_trace('Binascii exception - trying to decode string')
        _debug_print_trace(input_string)
        return None


def _unpack_and_hash_b64_binary(input_bytes, file_type=None):
    """
    If this is a known archive type extract the contents.

    Return a dictionary of (file_type :
        (file_name, file_hashes, input_bytes, decoded_string, encoding_type)
    """
    if not input_bytes:
        return None

    output_files = dict()
    if file_type in ['zip', 'gz', 'tar']:
        # if this is a known archive type - try to extract the contents
        (unpacked_type, file_items) = _get_items_from_archive(input_bytes, file_type)
        if unpacked_type != 'unknown':
            for file_name, extracted_file in file_items.items():
                file_results = _get_hashes_and_printable_string(extracted_file)
                idx = f'[{unpacked_type}] Filename: {file_name}'

                # ToDo - the unpacked type here refers to the archive file type so assigning this
                # to file_type is not exactly the right thing to do. In a future episode we'll
                # try to determine the file type using magic numbers.
                output_files[idx] = file_results._replace(file_name=file_name,
                                                          file_type=unpacked_type,
                                                          input_bytes=extracted_file)
                _debug_print_trace('_unpack_and_hash_b64_binary item (archive): ',
                                   type(file_results.decoded_string),
                                   file_results.decoded_string)

    if not output_files:
        # if this wasn't a known archive type or we failed to unpack anything,
        # just get the hashes and return
        file_results = _get_hashes_and_printable_string(input_bytes)
        idx = f'[{file_type}]'
        output_files[idx] = file_results._replace(file_name='unknown',
                                                  file_type=file_type,
                                                  input_bytes=input_bytes)
        _debug_print_trace('_unpack_and_hash_b64_binary item (other): ',
                           type(file_results.decoded_string),
                           file_results.decoded_string)
    return output_files


def _get_hashes_and_printable_string(extracted_file):
    """
    Get the hashes, encoding type and printable form of binary.

    (either string or list of hex-encoded byte values)
    """
    file_details = _get_byte_encoding(extracted_file)
    file_hashes = get_hashes(extracted_file)
    return file_details._replace(file_hashes=file_hashes)


def _get_items_from_archive(binary, archive_type='zip'):
    """Extract contained files from an archive type."""
    _debug_print_trace('_get_items_from_archive type: ', archive_type)
    if archive_type == 'zip':
        return get_items_from_zip(binary)
    elif archive_type == 'gz':
        return get_items_from_gzip(binary)
    elif archive_type == 'tar':
        return get_items_from_tar(binary)
    else:
        return 'unknown', {archive_type, binary}


@export
def get_items_from_gzip(binary):
    """Return decompressed gzip contents."""
    archive_file = gzip.decompress(binary)
    return 'gz', {'gzip_file': archive_file}


@export
def get_items_from_zip(binary):
    """Return dictionary of zip contents."""
    file_obj = io.BytesIO(binary)
    zip_archive = zipfile.ZipFile(file_obj, mode='r')
    archive_dict = dict()
    for item in zip_archive.namelist():
        archive_file = zip_archive.read(item)
        archive_dict[item] = archive_file
    return 'zip', archive_dict


@export
def get_items_from_tar(binary):
    """Return dictionary of tar file contents."""
    file_obj = io.BytesIO(binary)
    # Open tarfile
    tar = tarfile.open(mode="r", fileobj=file_obj)
    archive_dict = dict()
    # Iterate over every member
    for item in tar.getnames():
        tar_file = tar.extractfile(item)
        archive_file = tar_file.read()
        archive_dict[item] = archive_file
    return 'tar', archive_dict


@export
def get_hashes(binary):
    """Return md5, sha1 and sha256 hashes of input byte string."""
    hash_dict = dict()
    for hash_type in ['md5', 'sha1', 'sha256']:
        if hash_type == 'md5':
            hash_alg = hashlib.md5()
        elif hash_type == 'sha256':
            hash_alg = hashlib.sha256()
        elif hash_type == 'sha1':
            hash_alg = hashlib.sha1()
        hash_alg.update(binary)
        hash_dict[hash_type] = hash_alg.hexdigest()
    return hash_dict


def _binary_to_bytesio(binary):
    if isinstance(binary, io.BytesIO):
        return binary.getbuffer()
    return io.BytesIO(binary).getbuffer()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""query_mgr test class."""
import unittest
from datetime import datetime

from .. msticpy.nbtools import query_builtin_queries as queries
from .. msticpy.nbtools import kql
from .. msticpy.nbtools.query_mgr import replace_query_params, add_query, required_params
from .. msticpy.nbtools.query_defns import DataFamily, DataEnvironment, KqlQuery, QueryParamProvider


class QPTest(QueryParamProvider):
    """Unit test class."""
    def __init__(self, p_dict):
        self._p_dict = p_dict

    @property
    def query_params(self):
        """Return dict of query parameters."""
        return self._p_dict


class TestQueryManager(unittest.TestCase):
    """Unit test class."""

    def test_nbqueries(self):

        self.assertGreaterEqual(len(queries.query_definitions), 5)
        # try:
        #     subst_query = replace_query_params('list_alerts_counts')
        # except ValueError as val_err:
        self.assertRaises(ValueError, replace_query_params, 'list_alerts_counts')
        before_alerts = len(queries.query_definitions)
        add_query(KqlQuery(name='dummy_query',
                           query='''
                                {table}
                                {query_project}
                                | where StartTimeUtc >= datetime({start})
                                | where StartTimeUtc <= datetime({end})
                                | summarize alertCount=count(), firstAlert=min(StartTimeUtc),
                                lastAlert=max(StartTimeUtc) by AlertName
                                | order by alertCount desc
                                ''',
                           description='Retrieves summary of current alerts',
                           data_source='security_alert',
                           data_families=[DataFamily.WindowsSecurity],
                           data_environments=[DataEnvironment.LogAnalytics]))
        add_query(name='dummy_query2',
                  query='''
                        {table}
                        {query_project}
                        | where StartTimeUtc >= datetime({start})
                        | where StartTimeUtc <= datetime({end})
                        | summarize alertCount=count(), firstAlert=min(StartTimeUtc),
                        lastAlert=max(StartTimeUtc) by AlertName
                        | order by alertCount desc
                        ''',
                  description='Retrieves summary of current alerts',
                  data_source='security_alert',
                  data_families=[DataFamily.WindowsSecurity],
                  data_environments=[DataEnvironment.LogAnalytics])
        self.assertEqual(before_alerts + 2, len(queries.query_definitions))

        self.assertIn('get_alert', kql.__dict__)
        self.assertIn('get_process_parent', kql.__dict__)

        # This returns None and prints output but should execute with error
        self.assertRaises(LookupError, kql.get_process_parent)

        # from  datetime import datetime, timedelta
        # end_time = datetime.utcnow()
        # start_time = end_time - timedelta(hours=4)
        # try:
        #     nbqueries.list_alerts_counts(data_family='LinuxSecurity',
        # start=start_time, end=end_time)
        # except Exception as ex_type:
        #     pass

    def test_query_params(self):

        start_end = {
            'start': datetime(2018, 11, 23, 23, 36, 2, 894637),
            'end': datetime(2018, 11, 29, 23, 36, 2, 894637)}
        qptest1 = QPTest(start_end)

        self.assertIn('list_alerts_counts', queries.query_definitions)
        q_result = replace_query_params('list_alerts_counts', qptest1)
        self.assertIsNotNone(q_result)
        self.assertIn('SecurityAlert', q_result)
        self.assertIn('project', q_result)
        self.assertIn('2018-11-23', q_result)

        # Try with different query that expects an additional
        self.assertIn('get_alert', queries.query_definitions)
        try:
            q_result = replace_query_params('get_alert', qptest1)
        except:
            q_result = None
            pass
        self.assertIsNone(q_result)

        alertid = {'system_alert_id': '{some guid}'}

        qptest2 = QPTest(alertid)
        # Try with different query that expects an additional
        self.assertIn('get_alert', queries.query_definitions)
        q_result = replace_query_params('get_alert', qptest1, qptest2)
        self.assertIsNotNone(q_result)
        self.assertIn('SecurityAlert', q_result)
        self.assertIn('project', q_result)
        self.assertIn('2018-11-23', q_result)
        self.assertIn('SystemAlertId', q_result)
        self.assertIn('{some guid}', q_result)

        q_result3 = kql.get_alert(provs=[qptest1, qptest2])
        self.assertIsNotNone(q_result3)
        self.assertIn('SecurityAlert', q_result3)
        self.assertIn('project', q_result3)
        self.assertIn('2018-11-23', q_result3)
        self.assertIn('SystemAlertId', q_result3)
        self.assertIn('{some guid}', q_result3)

    def test_builtin_query_params(self):
        for _, kquery in queries.query_definitions.items():
            for param in required_params(kquery):
                self.assertIn(param, queries.KNOWN_PARAM_NAMES)


if __name__ == '__main__':
    unittest.main()
    print('bye')



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
"""security_alert test class."""
import sys
import unittest
import pandas as pd

from .. msticpy.nbtools import SecurityAlert
from .. msticpy.nbtools.security_alert_graph import create_alert_graph
from .. msticpy.nbtools.nbdisplay import display_alert
from .. msticpy.nbtools.query_defns import DataFamily, DataEnvironment


sample_alert = {
    'StartTimeUtc': '2018-09-27 16:59:16',
    'EndTimeUtc': '2018-09-27 16:59:16',
    'ProviderAlertId': 'b6329e79-0a94-4035-beee-c2e2657b71e3',
    'SystemAlertId': '2518642332435550951_b6329e79-0a94-4035-beee-c2',
    'ProviderName': 'Detection',
    'VendorName': 'Microsoft',
    'AlertType': 'RegistryPersistence',
    'AlertDisplayName': 'Windows registry persistence method detected',
    'Severity': 'Low',
    'IsIncident': False,
    'ExtendedProperties': {
        'resourceType': 'Non-Azure Resource',
        'enrichment_tas_threat__reports': '{"Kind":"MultiLink","DisplayValueToUrlDictionary":{"Report: Registry Persistence":"https://iflowreportsproda.blob.core.windows.net/reports/MSTI-TS-Registry-Persistence.pdf?sv=2018-03-28&sr=b&sig=TCqDRg1fQUp1P0gEznjM5wdqOGAKmAhsVzk4WPnx5gU%3D&spr=https&st=2018-09-27T13:37:57Z&se=2018-12-26T13:52:57Z&sp=r&callerId=ddd5443d-e6f4-441c-b52b-5278d2f21dfa&tenantId=bdb7e59d-8ba8-4351-aa6d-eedc6b916aa5"}}',
        'parent process id':  '0x940',
        'account logon id':  '0x3e7',
        'User SID':  'S-1-5-18',
        'command line':  'reg  add hkey_current_user\\software\\microsoft\\windows\\currentversion\\run /v cclientcleandll /t reg_sz /d "c:\\windows\\temp\\epdeldll.bat" /f',
        'process name':  'c:\\windows\\system32\\reg.exe',
        'process id':  '0x1670',
        'domain name':  'DOM',
        'user name':  'DOM\\TESTHOST$',
        'parent process':  'cmd.exe',
        'Persisted Process':  'c:\\windows\\temp\\epdeldll.bat'
    },
    'CompromisedEntity': 'TESTHOST',
    'Entities': [
        {'Type': 'host', '$id': '1', 'HostName': 'TESTHOST', 'DnsDomain': 'DOM.CONTOSO.COM', 'IsDomainJoined': True,
         'NTDomain': 'DOM', 'NetBiosName': 'TESTHOST', 'OsVersion': None, 'OSFamily': 'Windows'},
        {'Type': 'file', '$id': '2',
         'Directory': 'c:\\windows\\system32', 'Name': 'cmd.exe'},
        {'Type': 'process', '$id': '3', 'CommandLine': '', 'Host': {
            '$ref': '1'}, 'ProcessId': '0x940', 'ImageFile': {'$ref': '2'}},
        {'Type': 'account', '$id': '4', 'Name': 'TESTHOST$', 'IsDomainJoined': True,
         'NTDomain': 'DOM', 'Host': {'$ref': '1'}, 'LogonId': '0x3e7', 'Sid': 'S-1-5-18'},
        {'Type': 'file', '$id': '5',
         'Directory': 'c:\\windows\\system32', 'Name': 'reg.exe'},
        {'Type': 'process', '$id': '6', 'CommandLine': 'reg  add hkey_current_user\\software\\microsoft\\windows\\currentversion\\run /v cclientcleandll /t reg_sz /d "c:\\windows\\temp\\epdeldll.bat" /f',
         'Host': {'$ref': '1'}, 'ProcessId': '0x1670', 'ImageFile': {'$ref': '5'}, 'CreationTimeUtc': '2018-09-27T16:59:16.4449048Z', 'ParentProcess': {'$ref': '3'}, 'Account': {'$ref': '4'}, 'ElevationToken': 'Default'},
        {'Type': 'registry-key', '$id': '7',
         'Key': 'software\\microsoft\\windows\\currentversion\\run', 'Hive': 'HKEY_CURRENT_USER'},
        {'Type': 'registry-value', '$id': '8',
         'ValueType': 'Unknown', 'Key': {'$ref': '7'}},
        {'Type': 'registry-value', '$id': '9', 'Name': 'cclientcleandll', 'ValueType': 'String', 'Key': {'$ref': '7'}, 'Value': 'System.Byte[]'}],
    'ConfidenceLevel': 'Unknown',
    'ConfidenceScore': None,
    'ConfidenceReasons': None,
    'Intent': 'Persistence',
    'ExtendedLinks': None,
    'AzureResourceId': None,
    'AzureResourceSubscriptionId': None,
    'TenantId': 'b6329e79-0a94-4035-beee-c2e2657b71e3',
    'WorkspaceId': 'b6329e79-0a94-4035-beee-c2e2657b71e3',
    'AgentId': 'b6329e79-0a94-4035-beee-c2e2657b71e3',
    'SourceComputerId': 'b6329e79-0a94-4035-beee-c2e2657b71e3',
    'SystemSource': 'Non-Azure',
    'WorkspaceSubscriptionId': 'b6329e79-0a94-4035-beee-c2e2657b71e3',
    'WorkspaceResourceGroup': 'test-east-us',
    'TimeGeneratedUtc': '2018-09-27 16:59:47',
}


class TestQuerySchema(unittest.TestCase):

    def setUp(self):
        self.raw_alert = pd.Series(sample_alert)
        self.raw_alert['StartTimeUtc'] = pd.to_datetime(self.raw_alert['StartTimeUtc'])
        self.raw_alert['EndTimeUtc'] = pd.to_datetime(self.raw_alert['EndTimeUtc'])
        self.raw_alert['TimeGeneratedUtc'] = pd.to_datetime(self.raw_alert['TimeGeneratedUtc'])
    
    def test_alert_import(self):
        alert = SecurityAlert(self.raw_alert)

        str_alert = str(alert)
        self.assertIsNotNone(str_alert)
        self.assertIsNotNone(alert)
        self.assertGreaterEqual(len(alert.entities), 0)
        self.assertGreaterEqual(len(alert.ExtendedProperties), 0)
        self.assertGreaterEqual(len(alert['ExtendedProperties']), 0)
        self.assertIn('StartTimeUtc', alert)
        self.assertIn('EndTimeUtc', alert)
        self.assertIn('SystemAlertId', alert)
        self.assertIn('SystemAlertId', alert)
        self.assertIn('ProviderName', alert)
        self.assertIn('VendorName', alert)
        self.assertIn('AlertType', alert)
        self.assertIn('AlertDisplayName', alert)
        self.assertIn('Severity', alert)
        self.assertIn('IsIncident', alert)

        str_alert = str(alert)
        self.assertIsNotNone(str_alert)
        repr_alert = repr(alert)
        self.assertIsNotNone(repr_alert)

        #
        self.assertIsNotNone(alert.primary_host)
        self.assertEqual('TESTHOST', alert.primary_host.HostName)
        self.assertIsNotNone(alert.primary_process)
        self.assertIsNotNone(alert.primary_process.ProcessFilePath)
        self.assertEqual('c:\\windows\\system32\\reg.exe', alert.primary_process.ProcessFilePath)
        self.assertIsNotNone(alert.primary_account)
        self.assertEqual('TESTHOST$', alert.primary_account.Name)
        self.assertEqual('DOM\\TESTHOST$', alert.primary_account.qualified_name)
        self.assertEqual('0x3e7', alert.get_logon_id())

        self.assertIn('SourceComputerId', alert.host_filter(operator='=='))
        self.assertTrue(alert.is_in_log_analytics)
        self.assertTrue(alert.is_in_workspace)
        self.assertFalse(alert.is_in_azure_sub)
        self.assertIsNotNone(alert.host_filter(operator='=='))
        self.assertIn('true', alert.subscription_filter(operator='=='))

        self.assertEqual(2, len(alert.get_entities_of_type(entity_type='file')))

        self.assertGreater(len(alert.query_params), 5)
        self.assertEqual(alert.data_family, DataFamily.WindowsSecurity)
        self.assertEqual(alert.data_environment, DataEnvironment.LogAnalytics)

    def test_alert_display(self):

        alert = SecurityAlert(self.raw_alert)
        entity_str = ', '.join([str(e) for e in alert.entities])
        self.assertIsNotNone(entity_str)
        alert_html = alert.to_html(show_entities=True)
        self.assertIsNotNone(alert_html)
        alert_html = alert.to_html(show_entities=False)
        self.assertIsNotNone(alert_html)

    def test_alert_graph(self):
        alert = SecurityAlert(self.raw_alert)
        alert_graph = create_alert_graph(alert)
        self.assertIsNotNone(alert_graph)
        self.assertLessEqual(5, len(alert_graph.nodes))

    def test_alert_entities(self):
        alert = SecurityAlert(self.raw_alert)
        for ent in alert.entities:
            self.assertIsNotNone(ent.description_str)
